"""Unit Tests for Tools."""

# pylint: disable=redefined-outer-name
# pylint: disable=protected-access


# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from unittest.mock import patch, MagicMock
from dfcx_scrapi.core.tools import Tools
from google.cloud.dialogflowcx_v3beta1 import types
from google.cloud.dialogflowcx_v3beta1 import services

@pytest.fixture
def test_config():
    project_id = "my-project-id-1234"
    location_id = "global"
    parent = f"projects/{project_id}/locations/{location_id}"
    agent_id = f"{parent}/agents/my-agent-1234"
    tool_id = f"{agent_id}/tools/1234"
    display_name = "mock tool"
    description = "This is a mock tool."
    updated_description = "This is an updated mock tool."
    open_api_spec = """
    openapi: 3.0.0
    info:
    title: get_weather
    version: 1.0.0

    servers:
    - url: https://example.com

    paths:
    /get_weather_grid:
        get:
        summary: Returns the current grid information for a city and state
        operationId: get_weather_grid
        parameters:
            - name: latitude
            in: query
            required: true
            schema:
                type: string
            - name: longitude
            in: query
            required: true
            schema:
                type: string
        responses:
            '200':
            description: OK
            content:
                application/json:
                schema:
                    type: object
                    properties:
                    data:
                        type: string
    """

    return {
        "project_id": project_id,
        "agent_id": agent_id,
        "tool_id": tool_id,
        "display_name": display_name,
        "description": description,
        "updated_description": updated_description,
        "open_api_spec": open_api_spec
        }

@pytest.fixture
def open_api_tool(test_config):
    return types.Tool.OpenApiTool(
        text_schema=test_config["open_api_spec"]
    )

@pytest.fixture
def mock_tool_obj(test_config, open_api_tool):
    return types.Tool(
        name=test_config["tool_id"],
        display_name=test_config["display_name"],
        description=test_config["description"],
        open_api_spec=open_api_tool
        )

@pytest.fixture
def mock_tool_obj_updated(test_config, open_api_tool):
    return types.Tool(
        name=test_config["tool_id"],
        display_name=test_config["display_name"],
        description=test_config["updated_description"],
        open_api_spec=open_api_tool
        )

@pytest.fixture
def mock_list_tools_pager(mock_tool_obj):
    return services.tools.pagers.ListToolsPager(
        services.tools.ToolsClient.list_tools,
        types.tool.ListToolsRequest(),
        types.tool.ListToolsResponse(tools=[mock_tool_obj]),
    )

@pytest.fixture
def mock_list_playbooks_pager(mock_playbook_obj):
    return services.playbooks.pagers.ListPlaybooksPager(
        services.playbooks.PlaybooksClient.list_playbooks,
        types.playbook.ListPlaybooksRequest(),
        types.playbook.ListPlaybooksResponse(playbooks=[mock_playbook_obj]),
    )

@pytest.fixture(autouse=True)
def mock_client(test_config):
    """Fixture to create a mocked ToolsClient."""
    with patch("dfcx_scrapi.core.scrapi_base.default") as mock_default, \
        patch("dfcx_scrapi.core.scrapi_base.Request") as mock_request, \
        patch("dfcx_scrapi.core.tools.services.tools.ToolsClient") as mock_client:

        mock_creds = MagicMock()
        mock_default.return_value = (mock_creds, test_config["project_id"])
        mock_request.return_value = MagicMock()

        yield mock_client 

# Test get_tools_map
def test_get_tools_map(mock_client, mock_list_tools_pager, test_config):
    mock_client.return_value.list_tools.return_value = mock_list_tools_pager
    tools = Tools(agent_id=test_config["agent_id"])
    res = tools.get_tools_map(agent_id=test_config["agent_id"])

    assert isinstance(res, dict)
    assert test_config["tool_id"] in res
    assert res[test_config["tool_id"]] == test_config["display_name"]

# Test get_tools_map (reversed)
def test_get_tools_map_reversed(
    mock_client, mock_list_tools_pager, test_config):
    mock_client.return_value.list_tools.return_value = mock_list_tools_pager
    tools = Tools(agent_id=test_config["agent_id"])
    res = tools.get_tools_map(agent_id=test_config["agent_id"], reverse=True)

    assert isinstance(res, dict)
    assert test_config["display_name"] in res
    assert res[test_config["display_name"]] == test_config["tool_id"]

# Test list_tools
def test_list_tools(mock_client, mock_list_tools_pager, test_config):
    mock_client.return_value.list_tools.return_value = mock_list_tools_pager
    tools = Tools(agent_id=test_config["agent_id"])
    res = tools.list_tools(agent_id=test_config["agent_id"])

    assert isinstance(res, list)
    assert isinstance(res[0], types.Tool)

# Test get_tool
def test_get_tool(mock_client, mock_tool_obj, test_config):
    mock_client.return_value.get_tool.return_value = mock_tool_obj
    tools = Tools(agent_id=test_config["agent_id"])
    res = tools.get_tool(tool_id=test_config["tool_id"])

    assert isinstance(res, types.Tool)
    assert res.display_name == test_config["display_name"]

# Test create_tool
def test_create_tool_from_kwargs(
    mock_client, mock_tool_obj, test_config):
    mock_client.return_value.create_tool.return_value = mock_tool_obj
    tools = Tools(agent_id=test_config["agent_id"])
    res = tools.create_tool(
        agent_id=test_config["agent_id"],
        display_name=test_config["display_name"]
        )
    assert isinstance(res, types.Tool)
    assert res.display_name == test_config["display_name"]

def test_create_tool_from_proto_object(
    mock_client, mock_tool_obj, test_config):
    mock_client.return_value.create_tool.return_value = mock_tool_obj
    tools = Tools(agent_id=test_config["agent_id"])
    res = tools.create_tool(
        agent_id=test_config["agent_id"],
        obj=mock_tool_obj
        )
    assert isinstance(res, types.Tool)
    assert res.display_name == test_config["display_name"]

# Test delete_tool with tool_id
def test_delete_tool_with_tool_id(mock_client, test_config):
    tools = Tools(agent_id=test_config["agent_id"])
    tools.delete_tool(tool_id=test_config["tool_id"])
    mock_client.return_value.delete_tool.assert_called_once_with(
        name=test_config["tool_id"]
        )

# Test delete_tool with obj
def test_delete_tool_with_obj(mock_client, mock_tool_obj, test_config):
    tools = Tools(agent_id=test_config["agent_id"])
    tools.delete_tool(obj=mock_tool_obj)
    mock_client.return_value.delete_tool.assert_called_once_with(
        name=test_config["tool_id"]
        )

# Test update_tool with kwargs
def test_update_tool_with_kwargs(
    mock_client, mock_tool_obj_updated, test_config):
    mock_client.return_value.update_tool.return_value = mock_tool_obj_updated
    tools = Tools(agent_id=test_config["agent_id"])
    res = tools.update_tool(
        tool_id=test_config["tool_id"],
        description=test_config["updated_description"]
        )

    assert isinstance(res, types.Tool)
    assert res.description == test_config["updated_description"]

# Test building tool objects
def test_build_open_api_tool_no_description(test_config):
    tools = Tools(agent_id=test_config["agent_id"])
    tool = tools.build_open_api_tool(
        display_name=test_config["display_name"],
        spec=test_config["open_api_spec"],
    )

    assert isinstance(tool, types.Tool)
    assert tool.display_name == test_config["display_name"]
    assert tool.description == ""

def test_build_open_api_tool_with_description(test_config):
    tools = Tools(agent_id=test_config["agent_id"])
    tool = tools.build_open_api_tool(
        display_name=test_config["display_name"],
        spec=test_config["open_api_spec"],
        description=test_config["description"]
    )

    assert isinstance(tool, types.Tool)
    assert tool.display_name == test_config["display_name"]
    assert tool.description == test_config["description"]



"""Test Class for Agent Methods in SCRAPI."""

# pylint: disable=redefined-outer-name

# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from typing import Dict
from unittest.mock import patch, MagicMock
from google.protobuf import field_mask_pb2

from google.cloud.dialogflowcx_v3beta1 import types
from google.cloud.dialogflowcx_v3beta1.services.agents import (
    pagers, AgentsClient
    )

from dfcx_scrapi.core.agents import Agents

@pytest.fixture
def test_config():
    project_id = "my-project-id-1234"
    location_id = "global"
    parent = f"projects/{project_id}/locations/{location_id}"
    agent_id = f"{parent}/agents/my-agent-1234"
    display_name = "My Agent Display Name"
    default_id = "00000000-0000-0000-0000-000000000000"
    start_flow = f"{agent_id}/flows/{default_id}"
    start_playbook = f"{agent_id}/playbooks/{default_id}"
    return {
        "parent": parent,
        "project_id": project_id,
        "agent_id": agent_id,
        "display_name": display_name,
        "location_id": location_id,
        "start_flow": start_flow,
        "start_playbook": start_playbook
    }

@pytest.fixture
def mock_agent_obj(test_config: Dict[str, str]):
    return types.Agent(
        name=test_config["agent_id"],
        display_name=f"{test_config['display_name']}",
        default_language_code="en",
        time_zone="America/Chicago",
        start_flow=test_config["start_flow"]
    )

@pytest.fixture
def mock_agent_obj_playbook(test_config: Dict[str, str]):
    return types.Agent(
        name=test_config["agent_id"],
        display_name=test_config["display_name"],
        default_language_code="en",
        time_zone="America/Chicago",
        start_playbook=test_config["start_playbook"]
    )

@pytest.fixture
def mock_agent_obj_kwargs(mock_agent_obj: types.Agent):
    mock_agent_obj.description = "This is a Mock Agent description."
    mock_agent_obj.enable_stackdriver_logging = True

    return mock_agent_obj

@pytest.fixture
def mock_agent_obj_multiple_kwarg_changes(test_config: Dict[str, str]):
    return types.Agent(
        name=test_config["agent_id"],
        display_name="Updated Agent Display Name",
        default_language_code="es",
        time_zone="America/Chicago",
        start_flow=test_config["start_flow"]
    )

@pytest.fixture
def mock_updated_agent_obj(test_config: Dict[str, str]):
    return types.Agent(
        name=test_config["agent_id"],
        display_name="Updated Agent Display Name",
        default_language_code="en",
        time_zone="America/Chicago",
        start_flow=test_config["start_flow"]
    )

@pytest.fixture
def mock_list_agents_response(mock_agent_obj: types.Agent):
    return types.agent.ListAgentsResponse(agents=[mock_agent_obj])

@pytest.fixture
def mock_list_agents_pager(mock_list_agents_response: types.ListAgentsResponse):
    return pagers.ListAgentsPager(
        AgentsClient.list_agents,
        types.agent.ListAgentsRequest(),
        mock_list_agents_response,
    )

@pytest.fixture(autouse=True)
def mock_creds(test_config: Dict[str, str]):
    """Setup fixture for Agents Class to be used with all tests."""
    with patch("dfcx_scrapi.core.scrapi_base.default") as mock_default, \
        patch("dfcx_scrapi.core.scrapi_base.Request") as mock_request:
        mock_creds = MagicMock()
        mock_default.return_value = (mock_creds, test_config["project_id"])
        mock_request.return_value = MagicMock()

        yield mock_creds

@patch("dfcx_scrapi.core.agents.services.agents.AgentsClient.list_agents")
def test_list_agents_with_location(
        mock_list_agents: MagicMock,
        mock_list_agents_pager: pagers.ListAgentsPager,
        test_config: Dict[str, str]):
    mock_list_agents.return_value = mock_list_agents_pager
    expected_request = types.ListAgentsRequest(parent=test_config["parent"])

    agent = Agents()
    agents = agent.list_agents(
        project_id=test_config["project_id"],
        location_id="global"
    )
    assert isinstance(agents, list)
    assert isinstance(agents[0], types.agent.Agent)
    assert agents[0].name == test_config["agent_id"]
    mock_list_agents.assert_called_once_with(expected_request)

@patch("dfcx_scrapi.core.agents.services.agents.AgentsClient.list_agents")
def test_list_agents_without_location(
        mock_list_agents: MagicMock,
        mock_list_agents_pager: pagers.ListAgentsPager,
        test_config: Dict[str, str]):
    mock_list_agents.return_value = mock_list_agents_pager
    expected_regions = [
        "global", "us-central1", "us-east1", "us-west1", "asia-northeast1", 
        "asia-south1", "australia-southeast1", "northamerica-northeast1", 
        "europe-west1", "europe-west2"
    ]

    agent = Agents()
    agents = agent.list_agents(project_id=test_config["project_id"])
    assert isinstance(agents, list)
    assert isinstance(agents[0], types.agent.Agent)
    assert agents[0].name == test_config["agent_id"]
    assert mock_list_agents.call_count == 10
    for region in expected_regions:
        expected_request = types.ListAgentsRequest(
            parent=f"projects/{test_config['project_id']}/locations/{region}"
        )
        mock_list_agents.assert_any_call(expected_request)

@patch("dfcx_scrapi.core.agents.services.agents.AgentsClient.get_agent")
def test_get_agent(
        mock_get_agent: MagicMock,
        mock_agent_obj: types.Agent,
        test_config: Dict[str, str]):
    mock_get_agent.return_value = mock_agent_obj
    expected_request = types.GetAgentRequest(name=test_config["agent_id"])

    agent = Agents()
    response = agent.get_agent(test_config["agent_id"])

    assert isinstance(response, types.Agent)
    assert response.name == test_config["agent_id"]
    assert response.display_name == test_config["display_name"]
    mock_get_agent.assert_called_once_with(expected_request)

@patch("dfcx_scrapi.core.agents.services.agents.AgentsClient.list_agents")
def test_get_agent_by_display_name_no_location(
        mock_list_agents: MagicMock,
        test_config: Dict[str, str]):
    expected_regions = [
        "global", "us-central1", "us-east1", "us-west1", "asia-northeast1",
        "asia-south1", "australia-southeast1", "northamerica-northeast1",
        "europe-west1", "europe-west2"
    ]
    parent_stub = f"projects/{test_config['project_id']}/locations"
    agent_objects = [
        types.Agent(
            name=f"{parent_stub}/{region}/agents/my-agent-1234",
            display_name=f"{test_config['display_name']} {region}") 
        for region in expected_regions
    ]

    mock_list_agents.side_effect = [
        pagers.ListAgentsPager(
            AgentsClient.list_agents,
            types.agent.ListAgentsRequest(),
            types.agent.ListAgentsResponse(agents=[agent_obj]),
        ) for agent_obj in agent_objects
    ]

    agent = Agents()
    response = agent.get_agent_by_display_name(
        test_config["project_id"],
        f"{test_config['display_name']} global"
        )

    assert isinstance(response, types.Agent)
    assert response.name == test_config["agent_id"]
    assert response.display_name == f"{test_config['display_name']} global"
    assert mock_list_agents.call_count == 10
    for region in expected_regions:
        expected_request = types.ListAgentsRequest(
            parent=f"projects/{test_config['project_id']}/locations/{region}"
        )
        mock_list_agents.assert_any_call(expected_request)

@patch("dfcx_scrapi.core.agents.services.agents.AgentsClient.list_agents")
def test_get_agent_by_display_name_with_region(
        mock_list_agents: MagicMock,
        mock_list_agents_pager: pagers.ListAgentsPager,
        test_config: Dict[str, str]):
    mock_list_agents.return_value = mock_list_agents_pager
    expected_request = types.ListAgentsRequest(parent=test_config["parent"])

    agent = Agents()
    response = agent.get_agent_by_display_name(
        project_id=test_config["project_id"],
        display_name=test_config["display_name"],
        region="global"
        )

    assert isinstance(response, types.Agent)
    assert response.name == test_config["agent_id"]
    assert response.display_name == test_config["display_name"]
    mock_list_agents.assert_called_once_with(expected_request)

@patch("dfcx_scrapi.core.agents.services.agents.AgentsClient.create_agent")
def test_create_agent_with_kwargs(
        mock_create_agent: MagicMock,
        mock_agent_obj: types.Agent,
        test_config: Dict[str, str]):
    mock_create_agent.return_value = mock_agent_obj
    expected_agent = types.Agent(
        display_name=test_config["display_name"],
        default_language_code="en",
        time_zone="America/Chicago"
    )
    expected_request = types.CreateAgentRequest(
        parent=test_config["parent"],
        agent=expected_agent)

    agent = Agents()
    response = agent.create_agent(
        project_id=test_config["project_id"],
        display_name=test_config["display_name"]
    )

    assert isinstance(response, types.Agent)
    assert response.name == test_config["agent_id"]
    assert response.display_name == test_config["display_name"]
    mock_create_agent.assert_called_once_with(expected_request)

@patch("dfcx_scrapi.core.agents.services.agents.AgentsClient.create_agent")
def test_create_agent_from_obj(
        mock_create_agent: MagicMock,
        mock_agent_obj: types.Agent,
        test_config: Dict[str, str]):
    mock_create_agent.return_value = mock_agent_obj
    expected_agent = types.Agent(
        display_name=test_config["display_name"],
        default_language_code="en",
        time_zone="America/Chicago"
    )
    expected_request = types.CreateAgentRequest(
        parent=test_config["parent"],
        agent=expected_agent)

    agents = Agents()
    response = agents.create_agent(
        project_id=test_config["project_id"], obj=expected_agent)

    assert isinstance(response, types.Agent)
    assert response.name == test_config["agent_id"]
    assert response.display_name == test_config["display_name"]
    mock_create_agent.assert_called_once_with(expected_request)

@patch("dfcx_scrapi.core.agents.services.agents.AgentsClient.create_agent")
def test_create_agent_from_obj_with_kwargs(
        mock_create_agent: MagicMock,
        mock_agent_obj_kwargs: types.Agent,
        test_config):
    mock_create_agent.return_value = mock_agent_obj_kwargs
    expected_agent = types.Agent(
        display_name=test_config["display_name"],
        default_language_code="en",
        time_zone="America/Chicago",
        description="This is a Mock Agent description.",
        enable_stackdriver_logging=True,
    )
    expected_request = types.CreateAgentRequest(
        parent=test_config["parent"],
        agent=expected_agent)

    agents = Agents()
    response = agents.create_agent(
        project_id=test_config["project_id"],
        obj=expected_agent,
        description="This is a Mock Agent description.",
        enable_stackdriver_logging=True,
    )

    assert isinstance(response, types.Agent)
    assert response.name == test_config["agent_id"]
    assert response.description == "This is a Mock Agent description."
    mock_create_agent.assert_called_once_with(expected_request)

@patch("dfcx_scrapi.core.agents.services.agents.AgentsClient.update_agent")
def test_update_agent_with_obj_no_kwargs(
        mock_update_agent: MagicMock,
        mock_agent_obj: types.Agent,
        mock_updated_agent_obj: types.Agent,
        test_config: Dict[str, str]):
    """Test to ensure no updates are made without kwargs.

    This test will pass, but to the user it might seem like it will "fail" in
    the sense that the target Agent wasn't updated with their agent proto. This
    is because it is a _requirement_ to provide the exact kwargs you want to
    update on the Agent. If you only provide the object, there's no way to know
    exactly which items you want updated without explicitly passing them in
    kwargs.
    """
    mock_update_agent.return_value = mock_agent_obj

    # The request will be sent with the 'updated' obj the user provides
    # however, the actual response will be noop
    expected_request = types.UpdateAgentRequest(
        agent=mock_updated_agent_obj,
        update_mask=field_mask_pb2.FieldMask(paths=None)
        )

    agent = Agents()
    response = agent.update_agent(
        agent_id=test_config["agent_id"],
        obj=mock_updated_agent_obj
    )
    assert isinstance(response, types.Agent)
    assert response.name == test_config["agent_id"]
    assert response.display_name != "Updated Agent Display Name"
    mock_update_agent.assert_called_once_with(expected_request)

@patch("dfcx_scrapi.core.agents.services.agents.AgentsClient.update_agent")
def test_update_agent_with_obj_and_kwargs(
        mock_update_agent: MagicMock,
        mock_updated_agent_obj: types.Agent,
        mock_agent_obj_multiple_kwarg_changes: types.Agent,
        test_config: Dict[str, str]):
    """Test to ensure obj + kwargs results in updates.
    
    This test is a continuation of the test above, which check to ensure that
    when an obj + kwargs are provided, the kwargs are updated properly, and
    ONLY the kwargs provided are updated.
    """
    mock_update_agent.return_value = mock_updated_agent_obj

    # The user can send an object where multiple fields are changed but we will
    # only update the fields provided in the method input kwargs.
    expected_request = types.UpdateAgentRequest(
        agent=mock_agent_obj_multiple_kwarg_changes,
        update_mask=field_mask_pb2.FieldMask(paths=["display_name"])
        )

    agent = Agents()
    response = agent.update_agent(
        agent_id=test_config["agent_id"],
        obj=mock_agent_obj_multiple_kwarg_changes,
        display_name="Updated Agent Display Name"
    )

    assert isinstance(response, types.Agent)
    assert response.name == test_config["agent_id"]
    assert response.display_name == "Updated Agent Display Name"
    mock_update_agent.assert_called_once_with(expected_request)

    # Ensure lang code was not changed despite obj having changes
    assert response.default_language_code != "es"
    assert response.default_language_code == "en"

@patch("dfcx_scrapi.core.agents.services.agents.AgentsClient.get_agent")
@patch("dfcx_scrapi.core.agents.services.agents.AgentsClient.update_agent")
def test_update_agent_kwargs_only(
        mock_update_agent: MagicMock,
        mock_get_agent: MagicMock,
        mock_agent_obj: types.Agent,
        mock_updated_agent_obj: types.Agent,
        test_config: Dict[str, str]):
    mock_update_agent.return_value = mock_updated_agent_obj
    mock_get_agent.return_value = mock_agent_obj

    # Since the user doesn't provide the agent_obj, we'll also need to get the
    # agent for them before we update it, so we'll have 2 requests in this test
    get_request = types.GetAgentRequest(name=test_config["agent_id"])
    update_request = types.UpdateAgentRequest(
        agent=mock_updated_agent_obj,
        update_mask=field_mask_pb2.FieldMask(paths=["display_name"])
        )

    agent = Agents()
    response = agent.update_agent(
        agent_id=test_config["agent_id"],
        display_name="Updated Agent Display Name"
    )

    assert isinstance(response, types.Agent)
    assert response.name == test_config["agent_id"]
    assert response.display_name == "Updated Agent Display Name"
    mock_get_agent.assert_called_once_with(get_request)
    mock_update_agent.assert_called_once_with(update_request)

@patch("dfcx_scrapi.core.agents.services.agents.AgentsClient.delete_agent")
def test_delete_agent(
    mock_delete_agent: MagicMock,
    test_config: Dict[str, str]):
    res_str = f"Agent '{test_config['agent_id']}' successfully deleted."
    mock_delete_agent.return_value = res_str

    expected_request = types.DeleteAgentRequest(name=test_config["agent_id"])

    agent = Agents()
    response = agent.delete_agent(agent_id=test_config["agent_id"])

    assert response == res_str
    mock_delete_agent.assert_called_once_with(expected_request)



"""Base for other SCRAPI classes."""

# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import json
import re
import time
import functools
import threading
import vertexai
from collections import defaultdict
from typing import Dict, Any, Iterable, List

from google.auth import default
from google.api_core import exceptions
from google.cloud.dialogflowcx_v3beta1 import types
from google.oauth2 import service_account
from google.auth.transport.requests import Request
from google.protobuf import json_format
from google.protobuf import field_mask_pb2, struct_pb2

from vertexai.generative_models import GenerativeModel
from vertexai.language_models import TextEmbeddingModel, TextGenerationModel

from proto.marshal.collections import repeated
from proto.marshal.collections import maps

_INTERVAL_SENTINEL = object()

# The following models are supported for Metrics and Evaluations, either for
# Text Embeddings or used to provide Generations / Predictions.
SYS_INSTRUCT_MODELS = [
    "gemini-1.0-pro-002",
    "gemini-1.5-pro-001",
    "gemini-1.5-flash-001"
]

NON_SYS_INSTRUCT_GEM_MODELS = [
    "gemini-1.0-pro-001"
]

ALL_GEMINI_MODELS = SYS_INSTRUCT_MODELS + NON_SYS_INSTRUCT_GEM_MODELS

TEXT_GENERATION_MODELS = [
    "text-bison@002",
    "text-unicorn@001"
]
EMBEDDING_MODELS_NO_DIMENSIONALITY = [
    "textembedding-gecko@001",
    "textembedding-gecko@003",
    "textembedding-gecko-multilingual@001"
]
ALL_EMBEDDING_MODELS = EMBEDDING_MODELS_NO_DIMENSIONALITY + [
    "text-embedding-004"
]

ALL_GENERATIVE_MODELS = ALL_GEMINI_MODELS + TEXT_GENERATION_MODELS

# Define global scopes used for all Dialogflow CX Requests
GLOBAL_SCOPES = [
    "https://www.googleapis.com/auth/cloud-platform",
    "https://www.googleapis.com/auth/dialogflow",
    ]

class ScrapiBase:
    """Core Class for managing Auth and other shared functions."""

    def __init__(
        self,
        creds_path: str = None,
        creds_dict: Dict[str, str] = None,
        creds: service_account.Credentials = None,
        scope: List[str] = None,
        agent_id: str = None,
    ):

        self.scopes = GLOBAL_SCOPES
        if scope:
            self.scopes += scope

        if creds:
            self.creds = creds
            self.creds.refresh(Request())
            self.token = self.creds.token

        elif creds_path:
            self.creds = service_account.Credentials.from_service_account_file(
                creds_path, scopes=self.scopes
            )
            self.creds.refresh(Request())
            self.token = self.creds.token

        elif creds_dict:
            self.creds = service_account.Credentials.from_service_account_info(
                creds_dict, scopes=self.scopes
            )
            self.creds.refresh(Request())
            self.token = self.creds.token

        else:
            self.creds, _ = default()
            self.creds.refresh(Request())
            self.token = self.creds.token
            self._check_and_update_scopes(self.creds)

        self.agent_id = agent_id

        self.api_calls_dict = defaultdict(int)

    @staticmethod
    def _set_region(resource_id: str):
        """Different regions have different API endpoints

        Args:
          item_id: agent/flow/page - any type of long path id like
            `projects/<GCP PROJECT ID>/locations/<LOCATION ID>

        Returns:
          A dictionary containing the api_endpoint to use when
          instantiating other library client objects, or None
          if the location is "global"
        """
        try:
            location = resource_id.split("/")[3]
        except IndexError as err:
            logging.error("IndexError - path too short? %s", resource_id)
            raise err

        project_id = resource_id.split("/")[1]

        if location != "global":
            api_endpoint = f"{location}-dialogflow.googleapis.com:443"
            client_options = {
                "api_endpoint": api_endpoint,
                "quota_project_id": project_id}
            return client_options

        else:
            api_endpoint = "dialogflow.googleapis.com:443"
            client_options = {
                "api_endpoint": api_endpoint,
                "quota_project_id": project_id}

            return client_options

    @staticmethod
    def _client_options_discovery_engine(resource_id: str):
        """Different regions have different API endpoints

        Args:
          resource_id: any type of resource id associated with Discovery Engine
            resources. Ex: `projects/<GCP PROJECT ID>/locations/<LOCATION ID>`

        Returns:
          A dictionary containing the api_endpoint and quota project ID to use
          when calling Discovery Engine API endpoints.
        """
        try:
            location = resource_id.split("/")[3]
        except IndexError as err:
            logging.error("Please provide the fully qualified Resource ID: "
                          "%s", resource_id)
            raise err

        project_id = resource_id.split("/")[1]
        base_endpoint = "discoveryengine.googleapis.com:443"

        if location != "global":
            api_endpoint = f"{location}-{base_endpoint}"
            client_options = {
                "api_endpoint": api_endpoint,
                "quota_project_id": project_id}
            return client_options

        else:
            api_endpoint = base_endpoint
            client_options = {
                "api_endpoint": api_endpoint,
                "quota_project_id": project_id}

            return client_options

    @staticmethod
    def pbuf_to_dict(pbuf):
        """Extractor of json from a protobuf"""
        blobstr = json_format.MessageToJson(
            pbuf
        )
        blob = json.loads(blobstr)
        return blob

    @staticmethod
    def cx_object_to_json(cx_object):
        """Response objects have a magical _pb field attached"""
        return ScrapiBase.pbuf_to_dict(cx_object._pb)  # pylint: disable=W0212

    @staticmethod
    def cx_object_to_dict(cx_object):
        """Response objects have a magical _pb field attached"""
        return ScrapiBase.pbuf_to_dict(cx_object._pb)  # pylint: disable=W0212

    @staticmethod
    def extract_payload(msg):
        """Convert to json so we can get at the object"""
        blob = ScrapiBase.cx_object_to_dict(msg)
        return blob.get("payload")  # deref for nesting

    @staticmethod
    def dict_to_struct(some_dict: Dict[str, Any]):
        new_struct = struct_pb2.Struct()
        for k,v in some_dict.items():
            new_struct[k] = v

        return new_struct

    @staticmethod
    def parse_agent_id(resource_id: str):
        """Attempts to parse Agent ID from provided Resource ID."""
        parts = resource_id.split("/")
        if len(parts) < 6:
            raise ValueError(
                "Resource ID is too short to contain an Agent ID: {}".format(
                    resource_id
                )
            )

        if parts[4] != "agents":
            raise ValueError(
                "Resource ID does not contain an agent ID: {}".format(
                    resource_id
                )
            )

        return "/".join(parts[:6])

    @staticmethod
    def _parse_resource_path(
        resource_type,
        resource_id,
        validate=True) -> Dict[str, str]:
        # pylint: disable=line-too-long
        """Validates the provided Resource ID against known patterns.

        Args:
          resource_type, Must be one of the following resource types:
            `agent`, `entity_type`, `environmnet`, `flow`, `intent`, `page`,
            `project`, `security_setting`, `session`, `session_entity_type`,
            `test_case`, `transition_route_group`, `version`, `webhook`
          resource_id, The CX resource ID to check against the provided
            resource_type
          validate, allows the user to have their Resource ID validated along
            with returning the parts dictionary of the Resource ID. If set to
            True, this method will prompt the user with the correct format
            to utilize for the specified ID. If set to False, no validation
            will occur. If the input Resource ID is invalid when set to False,
            an empty Dictionary will be returned, allowing the caller to
            define their own ValueError message in a higher level class.
            Defaults to True.
        """

        data_store_match = r"[\w-]{1,947}"
        engine_match = r"[a-z0-9][a-z0-9-_]{0,62}"
        entity_id_match = r"[-@.0-9a-z]{1,36}"
        location_id_match = r"[-0-9a-z]{1,36}"
        page_id_match = r"[-0-9a-f]{1,36}|START_PAGE|END_SESSION|END_FLOW"
        session_id_match = r"[-0-9a-zA-Z!@#$%^&*()_+={}[\]:;\"'<>,.?]{1,36}"
        standard_id_match = r"[-0-9a-f]{1,36}"
        version_id_match = r"[0-9]{1,4}"

        matcher_root = f"^projects/(?P<project>.+?)/locations/(?P<location>{location_id_match})"

        pattern_map = {
            "agent": {
                "matcher": fr"{matcher_root}/agents/(?P<agent>{standard_id_match})$",
                "format": "`projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>`",
            },
            "data_store": {
                "matcher": fr"{matcher_root}/collections/default_collection/dataStores/(?P<data_store>{data_store_match})$",
                "format": "`projects/<Project ID>/locations/<Location ID>/collections/default_collection/dataStores/<Data Store ID>`"
            },
            "engine": {
                "matcher": fr"{matcher_root}/collections/default_collection/engines/(?P<engine>{engine_match})$",
                "format": "`projects/<Project ID>/locations/<Location ID>/collections/default_collection/engines/<Engine ID>`"
            },
            "entity_type": {
                "matcher": fr"{matcher_root}/agents/(?P<agent>{standard_id_match})/entityTypes/(?P<entity>{entity_id_match})$",
                "format": "`projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/entityTypes/<Entity Types ID>`",
            },
            "environment": {
                "matcher": fr"{matcher_root}/agents/(?P<agent>{standard_id_match})/environments/(?P<environment>{standard_id_match})$",
                "format": "`projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/environments/<Environment ID>`",
            },
            "flow": {
                "matcher": fr"{matcher_root}/agents/(?P<agent>{standard_id_match})/flows/(?P<flow>{standard_id_match})$",
                "format": "`projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/flows/<Flow ID>`",
            },
            "intent": {
                "matcher": fr"{matcher_root}/agents/(?P<agent>{standard_id_match})/intents/(?P<intent>{standard_id_match})$",
                "format": "`projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/intents/<Intent ID>`",
            },
            "page": {
                "matcher": fr"{matcher_root}/agents/(?P<agent>{standard_id_match})/flows/(?P<flow>{standard_id_match})/pages/(?P<page>{page_id_match})$",
                "format": "`projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/flows/<Flow ID>/pages/<Page ID>`",
            },
            "project": {
                "matcher": fr"{matcher_root}$",
                "format": "`projects/<Project ID>/locations/<Location ID>/`",
            },
            "security_setting": {
                "matcher": fr"{matcher_root}/securitySettings/(?P<security_setting>{standard_id_match})$",
                "format": "`projects/<Project ID>/locations/<Location ID>/securitySettings/<Security Setting ID>`",
            },
            "session": {
                "matcher": fr"{matcher_root}/agents/(?P<agent>{standard_id_match})/sessions/(?P<session>{session_id_match})$",
                "format": "`projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/sessions/<Session ID>`",
            },
            "session_entity_type": {
                "matcher": fr"{matcher_root}/agents/(?P<agent>{standard_id_match})/sessions/(?P<session>{session_id_match})/entityTypes/(?P<entity>{entity_id_match})$",
                "format": "`projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/sessions/<Session ID>/entityTypes/<Entity Type ID>`",
            },
            "test_case": {
                "matcher": fr"{matcher_root}/agents/(?P<agent>{standard_id_match})/testCases/(?P<test_case>{standard_id_match})$",
                "format": "`projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/testCases/<Test Case ID>`",
            },
            "transition_route_group": {
                "matcher": fr"{matcher_root}/agents/(?P<agent>{standard_id_match})/flows/(?P<flow>{standard_id_match})/transitionRouteGroups/(?P<transition_route_group>{standard_id_match})$",
                "format": "`projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/flows/<Flow ID>/transitionRouteGroups/<Transition Route Group ID>`",
            },
            "version": {
                "matcher": fr"{matcher_root}/agents/(?P<agent>{standard_id_match})/flows/(?P<flow>{standard_id_match})/versions/(?P<version>{version_id_match})$",
                "format": "`projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/flows/<Flow ID>`",
            },
            "webhook": {
                "matcher": fr"{matcher_root}/agents/(?P<agent>{standard_id_match})/webhooks/(?P<webhook>{standard_id_match})$",
                "format": "`projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>`",
            }
        }

        if resource_type not in pattern_map:
            raise KeyError(
                "`resource_type` must be one of the following resource types:"
                " `agent`, `data_store`, `engine`, `entity_type`, "
                "`environmnet`, `flow`, `intent`, `page`, `project`, "
                "`security_setting`, `session`, `session_entity_type`, "
                "`test_case`, `transition_route_group`, `version`, `webhook`"
            )

        match_res = re.match(pattern_map[resource_type]["matcher"], resource_id)
        dict_res = match_res.groupdict() if match_res else {}
        valid_parse = False

        if dict_res:
            valid_parse = True

        if validate and not valid_parse:
            raise ValueError(
                f"{resource_type.capitalize()} ID must be provided in the "
                f"following format: "
                f"{pattern_map[resource_type]['format']}"
            )

        # pylint: enable=line-too-long
        return dict_res

    @staticmethod
    def _validate_data_store_id(data_store_id: str):
        """Validate the data store ID and extract the ID if needed."""
        pattern = (r"^projects/(?P<project>.+?)/locations/"
                   r"(?P<location>[-0-9a-z]{1,36})/collections/"
                   r"default_collection/dataStores/"
                   r"(?P<data_store>[\w-]{1,947})$"
                   )
        match_res = re.match(pattern, data_store_id)
        if match_res:
            parts = match_res.groupdict()
            data_store_id = parts.get("data_store")

        return data_store_id

    @staticmethod
    def _get_solution_type(solution_type: str) -> int:
        """Get SOLUTION_TYPE from simple name reference."""
        solution_map = {
            "recommendation": 1,
            "search": 2,
            "chat": 3,
        }

        res = solution_map.get(solution_type, None)
        if not res:
            raise ValueError("Solution Type must be one of the following values"
                             ": `chat`, `search`, `recommendation`")

        return solution_map[solution_type]

    @staticmethod
    def is_valid_sys_instruct_model(llm_model: str) -> bool:
        valid_sys_instruct = True
        """Validate if model allows system instructions."""
        if llm_model in NON_SYS_INSTRUCT_GEM_MODELS:
            valid_sys_instruct = False

        return valid_sys_instruct

    def _check_and_update_scopes(self, creds: Any):
        """Update Credentials scopes if possible based on creds type."""
        if creds.requires_scopes:
            self.creds.scopes.extend(GLOBAL_SCOPES)

        else:
            logging.info("Found user creds, skipping global scopes...")

    def build_generative_model(
            self,
            llm_model: str,
            system_instructions: str = None
            ) -> GenerativeModel:
        """Build the GenertiveModel object and sys instructions as required."""
        valid_sys_intruct = self.is_valid_sys_instruct_model(llm_model)

        if valid_sys_intruct and system_instructions:
            return GenerativeModel(
                llm_model, system_instruction=system_instructions)

        elif not valid_sys_intruct and system_instructions:
            raise ValueError(
                f"Model `{llm_model}` does not support System Instructions"
                )
        else:
            return GenerativeModel(llm_model)

    def model_setup(self, llm_model: str, system_instructions: str = None):
        """Create a new LLM instance from user inputs."""
        if llm_model in ALL_EMBEDDING_MODELS:
            return TextEmbeddingModel.from_pretrained(llm_model)

        elif llm_model in ALL_GEMINI_MODELS:
            return self.build_generative_model(llm_model, system_instructions)

        elif llm_model in TEXT_GENERATION_MODELS:
            return TextGenerationModel.from_pretrained(llm_model)

        else:
            raise ValueError(f"LLM Model `{llm_model}` not supported.")

    def init_vertex(self, agent_id: str):
        """Use the Agent ID to parse out relevant fields and init Vertex API."""
        parts = self._parse_resource_path("agent", agent_id)
        project_id = parts.get("project")
        location = parts.get("location")

        # Vertex doesn't support global region at this time, but Agent Builder
        # and DFCX do. If this method is used in conjunction with an agent_id
        # that is located in `global` region, we will fallback to `us-central1`
        # to init Vertex.
        location = "us-central1" if location == "global" else location

        vertexai.init(project=project_id, location=location)

    def _build_data_store_parent(self, location: str) -> str:
        """Build the Parent ID needed for Discovery Engine API calls."""
        return (f"projects/{self.project_id}/locations/{location}/collections/"
                  "default_collection")

    def recurse_proto_repeated_composite(self, repeated_object):
        repeated_list = []
        for item in repeated_object:
            if isinstance(item, repeated.RepeatedComposite):
                item = self.recurse_proto_repeated_composite(item)
                repeated_list.append(item)
            elif isinstance(item, maps.MapComposite):
                item = self.recurse_proto_marshal_to_dict(item)
                repeated_list.append(item)
            else:
                repeated_list.append(item)

        return repeated_list

    def recurse_proto_marshal_to_dict(self, marshal_object):
        new_dict = {}
        for k, v in marshal_object.items():
            if isinstance(v, maps.MapComposite):
                v = self.recurse_proto_marshal_to_dict(v)
            elif isinstance(v, repeated.RepeatedComposite):
                v = self.recurse_proto_repeated_composite(v)
            new_dict[k] = v

        return new_dict

    def get_api_calls_details(self) -> Dict[str, int]:
        """The number of API calls corresponding to each method.

        Returns:
          A dictionary with keys as the method names
          and values as the number of calls.
        """
        this_class_methods, sub_class_apis_dict = {}, {}

        for attr_name in dir(self):
            attr = getattr(self, attr_name)
            if callable(attr) and hasattr(attr, "calls_api"):
                this_class_methods[attr_name] = 0
            if any(
                isinstance(attr, sub_class)
                for sub_class in ScrapiBase.__subclasses__()
            ):
                sub_class_apis_dict.update(attr.get_api_calls_details())

        if hasattr(self, "api_calls_dict"):
            this_class_methods.update(getattr(self, "api_calls_dict"))

        return {**this_class_methods, **sub_class_apis_dict}

    def get_api_calls_count(self) -> int:
        """Show the total number of API calls for this resource.

        Returns:
          Total calls to the API so far as an int.
        """
        return sum(self.get_api_calls_details().values())


    @staticmethod
    def _update_kwargs(obj, **kwargs) -> field_mask_pb2.FieldMask:
        """Create a FieldMask for Environment, Experiment, TestCase, Version."""
        if kwargs:
            for key, value in kwargs.items():
                setattr(obj, key, value)
            return field_mask_pb2.FieldMask(paths=kwargs.keys())
        attrs_map = {
            "Environment": [
                "name", "display_name", "description", "version_configs",
                "update_time", "test_cases_config", "webhook_config",
            ],
            "Experiment": [
                "name", "display_name", "description", "state", "definition",
                "rollout_config", "rollout_state", "rollout_failure_reason",
                "result", "create_time", "start_time", "end_time",
                "last_update_time", "experiment_length", "variants_history",
            ],
            "TestCase": [
                "name", "tags", "display_name", "notes", "test_config",
                "test_case_conversation_turns", "creation_time",
                "last_test_result",
            ],
            "Version": [
                "name", "display_name", "description", "nlu_settings",
                "create_time", "state",
            ],
        }
        if isinstance(obj, types.Environment):
            paths = attrs_map["Environment"]
        elif isinstance(obj, types.Experiment):
            paths = attrs_map["Experiment"]
        elif isinstance(obj, types.TestCase):
            paths = attrs_map["TestCase"]
        elif isinstance(obj, types.Version):
            paths = attrs_map["Version"]
        else:
            raise ValueError(
                "`obj` should be one of the following:"
                " [Environment, Experiment, TestCase, Version]."
            )

        return field_mask_pb2.FieldMask(paths=paths)


def api_call_counter_decorator(func):
    """Counts the number of API calls for the function `func`."""

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        self.api_calls_dict[func.__name__] += 1
        return func(self, *args, **kwargs)

    wrapper.calls_api = True

    return wrapper

def should_retry(err: exceptions.GoogleAPICallError) -> bool:
  """Helper function for deciding whether we should retry the error or not."""
  return isinstance(err, (exceptions.TooManyRequests, exceptions.ServerError))

def ratelimit(rate: float):
  """Decorator that controls the frequency of function calls."""
  seconds_per_event = 1.0 / rate
  lock = threading.Lock()
  bucket = 0
  last = 0

  def decorate(func):
    def rate_limited_function(*args, **kwargs):
      nonlocal last, bucket
      while True:
        with lock:
          now = time.time()
          bucket += now - last
          last = now

          # capping the bucket in order to avoid accumulating too many
          bucket = min(bucket, seconds_per_event)

          # if bucket is less than `seconds_per_event` then we have to wait
          # `seconds_per_event` - `bucket` seconds until a new "token" is
          # refilled
          delay = max(seconds_per_event - bucket, 0)

          if delay == 0:
            # consuming a token and breaking out of the delay loop to perform
            # the function call
            bucket -= seconds_per_event
            break
        time.sleep(delay)
      return func(*args, **kwargs)
    return rate_limited_function
  return decorate

def retry_api_call(retry_intervals: Iterable[float]):
  """Decorator for retrying certain GoogleAPICallError exception types."""
  def decorate(func):
    def retried_api_call_func(*args, **kwargs):
      interval_iterator = iter(retry_intervals)
      while True:
        try:
          return func(*args, **kwargs)
        except exceptions.GoogleAPICallError as err:
          print(f"retrying api call: {err}")
          if not should_retry(err):
            raise

          interval = next(interval_iterator, _INTERVAL_SENTINEL)
          if interval is _INTERVAL_SENTINEL:
            raise
          time.sleep(interval)
    return retried_api_call_func
  return decorate

def handle_api_error(func):
  """Decorator that chatches GoogleAPICallError exception and returns None."""
  def handled_api_error_func(*args, **kwargs):
    try:
      return func(*args, **kwargs)
    except exceptions.GoogleAPICallError as err:
      print(f"failed api call: {err}")
      return None

  return handled_api_error_func



"""Helper classes for parsing Agent Responses."""

# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
import json

from typing import Any, Union
from google.protobuf.json_format import MessageToDict
from google.cloud.dialogflowcx_v3beta1 import types

DataStoreConnectionSignals = (
    types.data_store_connection.DataStoreConnectionSignals
)

_EXECUTION_SEQUENCE_KEY = "DataStore Execution Sequence"

@dataclasses.dataclass
class Snippet:
   uri: Union[str, None]
   title: Union[str, None]
   text: Union[str, None]

   def to_prompt_snippet(self) -> str:
    result = []
    if self.title:
       result.append(self.title)
    if self.text:
        result.append(self.text)

    return "\n".join(result) if result else ""

@dataclasses.dataclass
class AgentResponse:
    """Dataclass for storing relevant fields of detect intent response."""
    # ResponseMessages
    answer_text: str = None

    # MatchType
    match_type: str = None

    # DataStoreConnectionSignals
    rewriter_llm_rendered_prompt: str = None
    rewriter_llm_output: str = None
    rewritten_query: str = None
    search_results: list[Snippet] = dataclasses.field(default_factory=list)
    answer_generator_llm_rendered_prompt: str = None
    answer_generator_llm_output: str = None
    generated_answer: str = None
    cited_snippet_indices: list[int] = dataclasses.field(default_factory=list)
    grounding_decision: str = None
    grounding_score: str = None
    safety_decision: str = None
    safety_banned_phrase_match: str = None

    # DiagnosticInfo ExecutionResult
    response_type: str = None
    response_reason: str = None
    latency: float = None
    faq_citation: bool = None
    search_fallback: bool = None
    unstructured_citation: bool = None
    website_citation: bool = None
    language: str = None

    def from_query_result(self, query_result: types.session.QueryResult):
        """Extracts the relevant fields from a QueryResult proto message."""
        answer_text = self._extract_text(query_result)
        match_type = self._extract_match_type(query_result)
        execution_result = self._extract_execution_result(query_result)

        self.answer_text=answer_text
        self.match_type=match_type
        self.response_type = execution_result.get("response_type")
        self.response_reason = execution_result.get("response_reason")
        self.latency = execution_result.get("latency")
        self.faq_citation = execution_result.get("faq_citation")
        self.search_fallback = execution_result.get("ucs_fallback")
        self.unstructured_citation = execution_result.get(
            "unstructured_citation")
        self.website_citation = execution_result.get("website_citation")
        self.language = execution_result.get("language")

        if query_result.data_store_connection_signals:
            self._extract_data_store_connection_signals(
                query_result.data_store_connection_signals
                )

    @classmethod
    def from_row(cls, row: dict[str, Any]):
        """Extracts the relevant fields from a dictionary."""
        row = row.copy()
        search_results = []
        for search_result in json.loads(row["search_results"]):
            search_results.append(Snippet(**search_result))

        row["search_results"] = search_results
        row["cited_snippet_indices"] = json.loads(row["cited_snippet_indices"])

        return cls(**row)

    def to_row(self):
        """Dumps the query result fields to a dictionary."""
        result = dataclasses.asdict(self)
        result["search_results"] = json.dumps(
            result.pop("search_results", []), indent=4
        )
        result["cited_snippet_indices"] = json.dumps(
            result["cited_snippet_indices"])

        return result

    @staticmethod
    def _extract_match_type(query_result: types.session.QueryResult) -> str:
        """Extracts the name of the match type from query result."""
        try:
            return types.session.Match.MatchType(
                query_result.match.match_type).name

        except ValueError:
            # if an enum type is returned which is not visible externally then
            # fallback to default value
            return types.session.Match.MatchType(0).name

    @staticmethod
    def _extract_text(res: types.session.QueryResult) -> str:
        all_text: list[str] = []
        if res.response_messages:
            for rm in res.response_messages:
                if rm.text and len(rm.text.text) > 0:
                    all_text.append(rm.text.text[0])

        final_text = "\n".join(all_text)

        return final_text

    @staticmethod
    def _extract_execution_result(
            query_result: types.session.QueryResult) -> dict[str, Any]:
        """Extracts the execution result from diagnostic info."""
        if _EXECUTION_SEQUENCE_KEY in query_result.diagnostic_info:
            execution_sequence = query_result.diagnostic_info[
                _EXECUTION_SEQUENCE_KEY
                ]
            if "executionResult" in execution_sequence:
                return MessageToDict(execution_sequence["executionResult"])
        return {}

    def _extract_search_results(
        self,
        data_store_connection_signals: DataStoreConnectionSignals
        ):
        """Extracts search results as a list of strings."""
        self.search_results = []
        for search_snippet in data_store_connection_signals.search_snippets:
            self.search_results.append(
                Snippet(
                    uri=search_snippet.document_uri,
                    title=search_snippet.document_title,
                    text=search_snippet.text,
                )
            )

    def _extract_citation_indices(
            self,
            data_store_connection_signals: DataStoreConnectionSignals
            ):
        """Extracts the links and snippets used to generate answer."""
        self.cited_snippet_indices = []
        for cited_snippet in data_store_connection_signals.cited_snippets:
            self.cited_snippet_indices.append(cited_snippet.snippet_index)

    @staticmethod
    def _extract_grounding_decision(
            grounding_signals: DataStoreConnectionSignals.GroundingSignals
            ) -> str:
        return DataStoreConnectionSignals.GroundingSignals.GroundingDecision(
            grounding_signals.decision
            ).name

    @staticmethod
    def _extract_grounding_score(
            grounding_signals: DataStoreConnectionSignals.GroundingSignals
            ):
        return DataStoreConnectionSignals.GroundingSignals.GroundingScoreBucket(
            grounding_signals.score
            ).name

    def _extract_grounding_signals(
            self, data_store_connection_signals: DataStoreConnectionSignals
            ) -> dict[str, str]:
        grounding_signals = data_store_connection_signals.grounding_signals
        if not grounding_signals:
            self.grounding_decision = None
            self.grounding_score = None
        else:
            self.grounding_decision = self._extract_grounding_decision(
                grounding_signals)
            self.grounding_score = self._extract_grounding_score(
                grounding_signals)

    def _extract_rewriter_llm_signals(
            self,
            data_store_connection_signals: DataStoreConnectionSignals
            ):
        rewriter_model_call_signals = (
            data_store_connection_signals.rewriter_model_call_signals
            )
        if not rewriter_model_call_signals:
            self.rewriter_llm_rendered_prompt = None
            self.rewriter_llm_output = None

        else:
            self.rewriter_llm_rendered_prompt = (
                rewriter_model_call_signals.rendered_prompt
            )
            self.rewriter_llm_output = rewriter_model_call_signals.model_output

    def _extract_answer_generator_llm_signals(
            self,
            data_store_connection_signals: DataStoreConnectionSignals
            ) -> dict[str, str]:
        answer_generation_model_call_signals = (
            data_store_connection_signals.answer_generation_model_call_signals
            )
        if not answer_generation_model_call_signals:
            self.answer_generator_llm_rendered_prompt = None
            self.answer_generator_llm_output = None

        else:
            self.answer_generator_llm_rendered_prompt = (
                answer_generation_model_call_signals.rendered_prompt
            )
            self.answer_generator_llm_output = (
                answer_generation_model_call_signals.model_output
            )

    @staticmethod
    def _extract_safety_decision(
            safety_signals: DataStoreConnectionSignals.SafetySignals) -> str:
        return DataStoreConnectionSignals.SafetySignals.SafetyDecision(
            safety_signals.decision
            ).name

    @staticmethod
    def _extract_safety_banned_phrase(
            safety_signals: DataStoreConnectionSignals.SafetySignals
            ) -> str:
        return DataStoreConnectionSignals.SafetySignals.BannedPhraseMatch(
            safety_signals.banned_phrase_match
            ).name

    def _extract_safety_signals(
            self, data_store_connection_signals: DataStoreConnectionSignals
            ) -> dict[str, str]:
        safety_signals = data_store_connection_signals.safety_signals
        if not safety_signals:
            self.safety_decision = None
            self.safety_banned_phrase_match = None
        else:
            self.safety_decision = self._extract_safety_decision(safety_signals)
            self.safety_banned_phrase_match = (
                self._extract_safety_banned_phrase(safety_signals)
            )

    def _extract_data_store_connection_signals(
            self,
            data_store_connection_signals: DataStoreConnectionSignals
            ) -> dict[str, Any]:
            self._extract_rewriter_llm_signals(data_store_connection_signals
                                               )
            self.rewritten_query = (
                data_store_connection_signals.rewritten_query
                if data_store_connection_signals.rewritten_query
                else None
            )

            self._extract_grounding_signals(data_store_connection_signals)
            self._extract_search_results(data_store_connection_signals)
            self._extract_answer_generator_llm_signals(
                data_store_connection_signals
                )
            self.generated_answer = (
                data_store_connection_signals.answer
                if data_store_connection_signals.answer
                else None
            )
            self._extract_citation_indices(data_store_connection_signals)
            self._extract_safety_signals(data_store_connection_signals)

    @property
    def search_result_links(self):
        return [search_result.uri for search_result in self.search_results]

    @property
    def cited_search_results(self):
        return [self.search_results[idx] for idx in self.cited_snippet_indices]

    @property
    def cited_search_result_links(self):
        return [search_result.uri for search_result in self.cited_search_results]

    @property
    def prompt_snippets(self):
        return [
            search_result.to_prompt_snippet()
            for search_result in self.search_results
        ]



"""Test Class for DialogflowConversation Methods in SCRAPI."""

# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import pandas as pd
from unittest.mock import patch, MagicMock
from google.oauth2.service_account import Credentials
from google.api_core.exceptions import InvalidArgument

from google.cloud.dialogflowcx_v3beta1 import types
from google.cloud.dialogflowcx_v3beta1 import services

from dfcx_scrapi.builders.flows import FlowBuilder
from dfcx_scrapi.builders.routes import TransitionRouteBuilder
from dfcx_scrapi.builders.fulfillments import FulfillmentBuilder
from dfcx_scrapi.builders.response_messages import ResponseMessageBuilder
from dfcx_scrapi.core.conversation import DialogflowConversation

@pytest.fixture
def test_config():
    project_id = "my-project-id-1234"
    location_id = "global"
    default_id = "00000000-0000-0000-0000-000000000000"
    another_id = "99999999-9999-9999-9999-999999999999"
    parent = f"projects/{project_id}/locations/{location_id}"
    agent_id = f"{parent}/agents/my-agent-1234"
    default_start_flow_id = f"{agent_id}/flows/{default_id}"
    another_flow_id = f"{agent_id}/flows/{another_id}"
    default_welcome_intent = f"{agent_id}/intents/{default_id}"
    email = "mock_email@testing.com"
    creds_path = "/Users/path/to/creds/credentials.json"
    creds_dict = {
        "type": "service_account",
        "project_id": project_id,
        "private_key_id": "1234",
        "private_key": "mock_key",
        "client_email": f"mock-account@{project_id}.iam.gserviceaccount.com",
        "client_id": "1234",
        "universe_domain": "googleapis.com",
    }

    mock_signer = MagicMock()
    mock_signer.key_id = "mock_key_id"
    mock_signer.sign.return_value = b"mock_signature"

    creds_object = Credentials(
        signer=mock_signer,
        token_uri="mock_token_uri",
        service_account_email=email,
        project_id=project_id,
        quota_project_id=project_id,
        scopes=[],
    )

    return {
        "project_id": project_id,
        "agent_id": agent_id,
        "default_welcome_intent": default_welcome_intent,
        "default_start_flow_id": default_start_flow_id,
        "another_flow_id": another_flow_id,
        "creds_path": creds_path,
        "creds_dict": creds_dict,
        "creds_object": creds_object,
    }

@pytest.fixture
def test_set_es():
    cols = ['flow_display_name', 'page_display_name', 'utterance']
    df = pd.DataFrame(columns=cols)
    df.loc[1] = ["Default Start Flow", "START_PAGE", "Hola!"]
    df.loc[2] = ["Default Start Flow", "START_PAGE", "Como estas?"]

    return df

@pytest.fixture
def test_set_en():
    cols = ['flow_display_name', 'page_display_name', 'utterance']
    df = pd.DataFrame(columns=cols)
    df.loc[1] = ["Default Start Flow", "START_PAGE", "Hi!"]
    df.loc[2] = ["Default Start Flow", "START_PAGE", "How are you?"]

    return df

@pytest.fixture
def es_response_message():
    rmb = ResponseMessageBuilder()
    msg = rmb.create_new_proto_obj(
        response_type="text",
        message=["¡Hola!", "¡Buenos días!"]
    )

    return msg

@pytest.fixture
def nlu_settings():
    fb = FlowBuilder()
    flow = fb.create_new_proto_obj(
        display_name="Default Start Flow",
        overwrite=True
    )
    flow = fb.nlu_settings(
        model_type=3,
        classification_threshold=0.3,
        model_training_mode=1
    )

    return flow.nlu_settings

@pytest.fixture
def mock_flows_list_es(test_config, nlu_settings, es_response_message):
    fulb = FulfillmentBuilder()
    trb = TransitionRouteBuilder()

    agent_msg = fulb.create_new_proto_obj()
    agent_msg = fulb.add_response_message(es_response_message)

    route = trb.create_new_proto_obj(
        intent=test_config["default_welcome_intent"],
        trigger_fulfillment=agent_msg
    )

    flow1 = types.Flow(
        name = test_config["default_start_flow_id"],
        display_name = "Default Start Flow",
        description = "Basic Flow in Spanish",
        transition_routes = [route],
        nlu_settings = nlu_settings
    )

    flow2 = types.Flow(
        name = test_config["another_flow_id"],
        display_name = "Flow #2",
        description = "Another Flow in Spanish",
        transition_routes = [route],
        nlu_settings = nlu_settings
    )

    return [flow1, flow2]


@pytest.fixture
def mock_pages_list(test_config):
    flow_id1 = test_config["default_start_flow_id"]
    flow_id2 = test_config["another_flow_id"]

    page1 = types.Page(
        name = f"{flow_id1}/pages/1234",
        display_name = "Test Page 1"
    )

    page2 = types.Page(
        name = f"{flow_id2}/pages/9876",
        display_name = "Test Page 2"
    )

    return [page1, page2]

@pytest.fixture
def mock_list_flows_pager(mock_flows_list_es):
    return services.flows.pagers.ListFlowsPager(
        services.flows.FlowsClient.list_flows,
        types.ListFlowsRequest(),
        types.ListFlowsResponse(
            flows=mock_flows_list_es),
    )

@pytest.fixture
def mock_list_pages_pager(mock_pages_list):
    return services.pages.pagers.ListPagesPager(
        services.pages.PagesClient.list_pages,
        types.ListPagesRequest(),
        types.ListPagesResponse(
            pages=mock_pages_list),
    )

@pytest.fixture(autouse=True)
def mock_client(test_config):
    """Fixture to create a mocked DialogflowConversation client."""
    with patch("dfcx_scrapi.core.scrapi_base.default") as mock_default, \
        patch("dfcx_scrapi.core.scrapi_base.Request") as mock_request, \
        patch("dfcx_scrapi.core.conversation.DialogflowConversation") as mock_client:

        mock_creds = MagicMock()
        mock_default.return_value = (mock_creds, test_config["project_id"])
        mock_request.return_value = MagicMock()

        yield mock_client

@patch("dfcx_scrapi.core.flows.services.flows.FlowsClient.list_flows")
@patch("dfcx_scrapi.core.pages.pages.PagesClient.list_pages")
def test_page_id_mapper_with_config_lang_code(
    mock_list_pages,
    mock_list_flows,
    test_config,
    mock_list_flows_pager,
    mock_list_pages_pager
    ):
    mock_list_pages.return_value = mock_list_pages_pager
    mock_list_flows.return_value = mock_list_flows_pager
    expected_request = types.flow.ListFlowsRequest(
        parent=test_config["agent_id"], language_code="es"
    )

    config = {
        "agent_path": test_config["agent_id"],
        "language_code": "es"
    }

    dc = DialogflowConversation(config=config)
    dc._page_id_mapper()

    assert dc.agent_id == test_config["agent_id"]
    assert dc.language_code == "es"
    assert dc.flows.language_code == "es"
    assert dc.pages.language_code == "es"
    assert "Default Start Flow" in dc.agent_pages_map.flow_display_name.unique()
    assert "Flow #2" in dc.agent_pages_map.flow_display_name.unique()
    mock_list_flows.assert_called_once_with(expected_request)

@patch("dfcx_scrapi.core.flows.services.flows.FlowsClient.list_flows")
@patch("dfcx_scrapi.core.pages.pages.PagesClient.list_pages")
def test_page_id_mapper_with_class_lang_code(
    mock_list_pages,
    mock_list_flows,
    test_config,
    mock_list_flows_pager,
    mock_list_pages_pager
    ):
    mock_list_pages.return_value = mock_list_pages_pager
    mock_list_flows.return_value = mock_list_flows_pager
    expected_request = types.flow.ListFlowsRequest(
        parent=test_config["agent_id"], language_code="es"
    )

    config = {"agent_path": test_config["agent_id"]}

    dc = DialogflowConversation(config=config, language_code="es")
    dc._page_id_mapper()

    assert dc.agent_id == test_config["agent_id"]
    assert dc.language_code == "es"
    assert dc.flows.language_code == "es"
    assert dc.pages.language_code == "es"
    assert "Default Start Flow" in dc.agent_pages_map.flow_display_name.unique()
    assert "Flow #2" in dc.agent_pages_map.flow_display_name.unique()
    mock_list_flows.assert_called_once_with(expected_request)

@patch("dfcx_scrapi.core.flows.services.flows.FlowsClient.list_flows")
@patch("dfcx_scrapi.core.pages.pages.PagesClient.list_pages")
def test_page_id_mapper_with_default_lang_code(
    mock_list_pages,
    mock_list_flows,
    test_config,
    mock_list_flows_pager,
    mock_list_pages_pager
    ):
    mock_list_pages.return_value = mock_list_pages_pager
    mock_list_flows.return_value = mock_list_flows_pager
    expected_request = types.flow.ListFlowsRequest(
        parent=test_config["agent_id"], language_code="en"
    )

    config = {"agent_path": test_config["agent_id"]}

    dc = DialogflowConversation(config=config)
    dc._page_id_mapper()

    assert dc.agent_id == test_config["agent_id"]
    assert dc.language_code == "en"
    assert dc.flows.language_code == "en"
    assert dc.pages.language_code == "en"
    assert "Default Start Flow" in dc.agent_pages_map.flow_display_name.unique()
    assert "Flow #2" in dc.agent_pages_map.flow_display_name.unique()
    mock_list_flows.assert_called_once_with(expected_request)

@patch("dfcx_scrapi.core.flows.services.flows.FlowsClient.list_flows")
@patch("dfcx_scrapi.core.agents.Agents.get_agent")
def test_page_id_mapper_config_invalid_lang_code(
    mock_get_agent,
    mock_list_flows,
    test_config
    ):
    mock_agent = types.Agent(supported_language_codes=["es", "fr"])
    mock_get_agent.return_value = mock_agent
    expected_request = types.flow.ListFlowsRequest(
        parent=test_config["agent_id"], language_code="en"
    )
    
    # Mock the list_flows method to raise an InvalidArgument exception
    # with the specific error message you want to test for.
    mock_list_flows.side_effect = InvalidArgument(
        "400 com.google.apps.framework.request.BadRequestException: "
        "Agent does not support language: 'en'."
    )

    config = {
        "agent_path": test_config["agent_id"],
        "language_code": "en"
    }

    with pytest.raises(InvalidArgument) as excinfo:
        dc = DialogflowConversation(config=config)
        dc._page_id_mapper()

    assert "Agent does not support language: 'en'." in str(excinfo.value)
    mock_list_flows.assert_called_once_with(expected_request)

@patch("dfcx_scrapi.core.flows.services.flows.FlowsClient.list_flows")
@patch("dfcx_scrapi.core.agents.Agents.get_agent")
def test_page_id_mapper_class_invalid_lang_code(
    mock_get_agent,
    mock_list_flows,
    test_config
    ):
    mock_agent = types.Agent(supported_language_codes=["es", "fr"])
    mock_get_agent.return_value = mock_agent
    expected_request = types.flow.ListFlowsRequest(
        parent=test_config["agent_id"], language_code="en"
    )
    
    # Mock the list_flows method to raise an InvalidArgument exception
    # with the specific error message you want to test for.
    mock_list_flows.side_effect = InvalidArgument(
        "400 com.google.apps.framework.request.BadRequestException: "
        "Agent does not support language: 'en'."
    )

    config = {
        "agent_path": test_config["agent_id"],
    }

    with pytest.raises(InvalidArgument) as excinfo:
        dc = DialogflowConversation(config=config, language_code="en")
        dc._page_id_mapper()

    assert "Agent does not support language: 'en'." in str(excinfo.value)
    mock_list_flows.assert_called_once_with(expected_request)



"""CX Session Resource functions."""

# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import uuid
from typing import Dict, List
from google.cloud.dialogflowcx_v3beta1 import services
from google.cloud.dialogflowcx_v3beta1 import types
from google.protobuf.json_format import MessageToDict
from proto.marshal.collections import maps
from IPython.display import display, Markdown

from dfcx_scrapi.core import scrapi_base
from dfcx_scrapi.core import tools
from dfcx_scrapi.core import playbooks
from dfcx_scrapi.core import flows

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class Sessions(scrapi_base.ScrapiBase):
    """Core Class for CX Session Resource functions."""

    def __init__(
        self,
        creds_path: str = None,
        creds_dict: Dict = None,
        creds=None,
        scope=False,
        agent_id: str = None,
        session_id: str = None,
        tools_map: Dict[str, str] = None,
        playbooks_map: Dict[str, str] = None,
        flows_map: Dict[str, str] = None
    ):
        super().__init__(
            creds_path=creds_path, creds_dict=creds_dict,
            creds=creds, scope=scope
        )

        self.session_id = session_id
        self.agent_id = agent_id
        self.tools_map = tools_map
        self.playbooks_map = playbooks_map
        self.flows_map = flows_map

    @property
    def session_id(self):
        return self._session_id

    @session_id.setter
    def session_id(self, value):
        if value:
            self._parse_resource_path("session", value)

        self._session_id = value

    @staticmethod
    def printmd(string):
        display(Markdown(string))

    @staticmethod
    def _build_query_input(text, language_code):
        """Build out the query_input object for the Query Request.

        Args:
          text, the text to use for the Detect Intent request.
          language_code, the language code to use for Detect Intent request.
        """
        text_input = types.session.TextInput(text=text)
        query_input = types.session.QueryInput(
            text=text_input, language_code=language_code
        )

        return query_input

    @staticmethod
    def build_intent_query_input(intent_id: str, language_code: str):
        """Build the query_input object for direct Intent request."""
        intent_input = types.session.IntentInput(intent=intent_id)
        query_input = types.session.QueryInput(
            intent=intent_input, language_code=language_code
        )

        return query_input

    @staticmethod
    def get_tool_action(tool_use: types.example.ToolUse) -> str:
        return tool_use.action

    def get_tool_params(self, params: maps.MapComposite):
        "Handle various types of param values from Tool input/outputs."
        param_map = {}
        if isinstance(params, maps.MapComposite):
            param_map = self.recurse_proto_marshal_to_dict(params)

        # Clean up resulting param map. This is because I/O params from Agent
        # Builder proto will have a blank top level key, but the main value
        # info is what is important for return to the user in this tool.
        empty_top_key = param_map.get("", None)
        if len(param_map.keys()) == 1 and empty_top_key:
            param_map = param_map[""]

        return param_map

    def get_playbook_name(self, playbook_id: str):
        agent_id = self.parse_agent_id(playbook_id)
        if not self.playbooks_map:
            playbook_client = playbooks.Playbooks(
                agent_id=agent_id, creds=self.creds
                )
            self.playbooks_map = playbook_client.get_playbooks_map(agent_id)

        return self.playbooks_map[playbook_id]

    def get_tool_name(self, tool_use: types.example.ToolUse) -> str:
        agent_id = self.parse_agent_id(tool_use.tool)
        if not self.tools_map:
            tool_client = tools.Tools(creds=self.creds)
            self.tools_map = tool_client.get_tools_map(agent_id)
        return self.tools_map[tool_use.tool]

    def get_flow_name(self, flow_id: str):
        agent_id = self.parse_agent_id(flow_id)
        if not self.flows_map:
            flow_client = flows.Flows(creds=self.creds)
            self.flows_map = flow_client.get_flows_map(agent_id)

        return self.flows_map[flow_id]

    def collect_tool_responses(
        self, res: types.session.QueryResult
    ) -> List[Dict[str, str]]:
        """Gather all the tool responses into a list of dicts."""
        tool_responses = []
        for action in res.generative_info.action_tracing_info.actions:
            if action.tool_use:
                tool_responses.append(
                    {
                        "tool_name": self.get_tool_name(action.tool_use),
                        "tool_action": self.get_tool_action(action.tool_use),
                        "input_params": self.get_tool_params(
                            action.tool_use.input_action_parameters),
                        "output_params": self.get_tool_params(
                            action.tool_use.output_action_parameters),
                    }
                )

        return tool_responses

    def collect_playbook_responses(
        self, res: types.session.QueryResult
    ) -> List[Dict[str, str]]:
        """Gather all the playbook responses into a list of dicts."""
        playbook_responses = []
        for action in res.generative_info.action_tracing_info.actions:
            if action.playbook_invocation:
                playbook_responses.append(
                    {
                        "playbook_name": self.get_playbook_name(
                            action.playbook_invocation.playbook
                        )
                    }
                )
            # If no playbook invocation was found try to return the current
            # Playbook
            elif len(res.generative_info.current_playbooks) > 0:
                playbook_responses.append(
                    {
                        "playbook_name": self.get_playbook_name(
                            res.generative_info.current_playbooks[-1]
                            )
                    }
                )

        return playbook_responses

    def collect_flow_responses(
        self, res: types.session.QueryResult
    ) -> List[Dict[str, str]]:
        """Gather all the flow repsonses into a list of dicts."""
        flow_responses = []
        for action in res.generative_info.action_tracing_info.actions:
            if action.flow_invocation:
                flow_responses.append(
                    {
                        "flow_name": self.get_flow_name(
                            action.flow_invocation.flow
                        )
                    }
                )

        return flow_responses

    def build_session_id(
        self, agent_id: str = None, overwrite: bool = True
    ) -> str:
        """Creates a valid UUID-4 Session ID to use with other methods.

        Args:
          overwrite (Optional), if a session_id already exists, this will
            overwrite the existing Session ID parameter. Defaults to True.
        """

        agent_parts = self._parse_resource_path("agent", agent_id)
        session_id = (
            f"projects/{agent_parts['project']}/"
            f"locations/{agent_parts['location']}/agents/"
            f"{agent_parts['agent']}/sessions/{uuid.uuid4()}"
        )

        if overwrite:
            self.session_id = session_id

        return session_id

    def run_conversation(
        self,
        agent_id: str = None,
        session_id: str = None,
        conversation: List[str] = None,
        parameters=None,
        response_text=False,
    ):
        """Tests a full conversation with the specified CX Agent.

        Args:
          agent_id: the Agent ID of the CX Agent to have the conversation with.
          session_id: an RFC 4122 formatted UUID to be used as the unique ID
            for the duration of the conversation session. When using Python
            uuid library, uuid.uuid4() is preferred.
          conversation: a List of Strings that represent the USER utterances
            for the given conversation, in the order they would happen
            chronologically in the conversation.
            Ex:
              ['I want to check my bill', 'yes', 'no that is all', 'thanks!']
          parameters: (Optional) Dict of CX Session Parameters to set in the
            conversation. Typically this is set before a conversation starts.
          response_text: Will provide the Agent Response text if set to True.
            Default value is False.

        Returns:
          None, the conversation Request/Response is printed to console.
        """
        if not session_id:
            session_id = self.session_id

        client_options = self._set_region(agent_id)
        session_client = services.sessions.SessionsClient(
            client_options=client_options, credentials=self.creds
        )
        session_path = f"{agent_id}/sessions/{session_id}"

        if parameters:
            query_params = types.session.QueryParameters(parameters=parameters)
            text_input = types.session.TextInput(text="")
            query_input = types.session.QueryInput(
                text=text_input, language_code="en"
            )
            request = types.session.DetectIntentRequest(
                session=session_path,
                query_params=query_params,
                query_input=query_input,
            )

            response = session_client.detect_intent(request=request)

        for text in conversation:
            text_input = types.session.TextInput(text=text)
            query_input = types.session.QueryInput(
                text=text_input, language_code="en"
            )
            request = types.session.DetectIntentRequest(
                session=session_path, query_input=query_input
            )
            response = session_client.detect_intent(request=request)
            query_result = response.query_result

            print("=" * 20)
            print(f"Query text: {query_result.text}")
            if "intent" in query_result:
                print(f"Triggered Intent: {query_result.intent.display_name}")

            if "intent_detection_confidence" in query_result:
                print(
                    f"Intent Confidence: \
                        f{query_result.intent_detection_confidence}"
                )

            print(f"Response Page: {query_result.current_page.display_name}")

            for param in query_result.parameters:
                if param == "statusMessage":
                    print(f"Status Message: {query_result.parameters[param]}")

            if response_text:
                concat_messages = " ".join(
                    [
                        " ".join(response_message.text.text)
                        for response_message in query_result.response_messages
                    ]
                )
                print(f"Response Text: {concat_messages}\n")

    def detect_intent(
        self,
        agent_id,
        session_id,
        text,
        language_code="en",
        parameters=None,
        end_user_metadata=None,
        populate_data_store_connection_signals=False,
        intent_id: str = None
    ):
        """Returns the result of detect intent with texts as inputs.

        Using the same `session_id` between requests allows continuation
        of the conversation.

        Args:
          agent_id: the Agent ID of the CX Agent to have the conversation with.
          session_id: an RFC 4122 formatted UUID to be used as the unique ID
            for the duration of the conversation session. When using Python
            uuid library, uuid.uuid4() is preferred.
          text: the user utterance to run intent detection on
          parameters: (Optional) Dict of CX Session Parameters to set in the
            conversation. Typically this is set before a conversation starts.
          end_user_metadata: (Optional) Dict of CX Session endUserMetadata to
            set in the conversation.
          populate_data_store_connection_signals: If set to true and data
            stores are involved in serving the request then query result will
            be populated with data_store_connection_signals field which
            contains data that can help evaluations.

        Returns:
          The CX query result from intent detection
        """
        client_options = self._set_region(agent_id)
        session_client = services.sessions.SessionsClient(
            client_options=client_options, credentials=self.creds
        )

        res = self._parse_resource_path("session", str(session_id), False)
        if not res:
            raise ValueError(
                "Session ID must be provided in the following format: "
                "`projects/<Project ID>/locations/<Location ID>/agents/"
                "<Agent ID>/sessions/<Session ID>`.\n\n"
                "Utilize `build_session_id` to create a new Session ID."
            )

        if intent_id:
            query_input = self.build_intent_query_input(
                intent_id, language_code)
        else:
            query_input = self._build_query_input(text, language_code)

        request = types.session.DetectIntentRequest()
        request.session = session_id
        request.query_input = query_input

        query_param_mapping = {}

        if parameters:
            query_param_mapping["parameters"] = parameters

        if end_user_metadata:
            query_param_mapping["end_user_metadata"] = end_user_metadata

        if populate_data_store_connection_signals:
            query_param_mapping[
                "populate_data_store_connection_signals"
            ] = populate_data_store_connection_signals

        if query_param_mapping:
            query_params = types.session.QueryParameters(query_param_mapping)
            request.query_params = query_params

        response = session_client.detect_intent(request)
        query_result = response.query_result

        return query_result

    def preset_parameters(
        self, agent_id: str = None, session_id: str = None, parameters=None
    ):
        """Used to set session parameters before a conversation starts.

        Args:
          agent_id: the Agent ID of the CX Agent to have the conversation with.
          session_id: an RFC 4122 formatted UUID to be used as the unique ID
            for the duration of the conversation session. When using Python
            uuid library, uuid.uuid4() is preferred.
          parameters: Dict of CX Session Parameters to set in the
            conversation. Typically this is set before a conversation starts.

        Returns:
          The CX query result from intent detection run with no text input
        """
        client_options = self._set_region(agent_id)
        session_client = services.sessions.SessionsClient(
            client_options=client_options, credentials=self.creds
        )
        session_path = f"{agent_id}/sessions/{session_id}"

        query_params = types.session.QueryParameters(parameters=parameters)
        text_input = types.session.TextInput(text=None)
        query_input = types.session.QueryInput(
            text=text_input, language_code="en"
        )
        request = types.session.DetectIntentRequest(
            session=session_path,
            query_params=query_params,
            query_input=query_input,
        )

        response = session_client.detect_intent(request=request)

        return response

    def get_agent_answer(self, user_query: str) -> str:
        """Extract the answer/citation from a Vertex Conversation response."""

        session_id = self.build_session_id(self.agent_id)
        res = MessageToDict(
            self.detect_intent(  # pylint: disable=W0212
                self.agent_id, session_id, user_query
            )._pb
        )

        answer_text = res["responseMessages"][0]["text"]["text"][0]
        answer_link = (
            res["responseMessages"][1]["payload"]["richContent"][0][0][
                "actionLink"
            ]
            if len(res["responseMessages"]) > 1
            else ""
        )

        return f"{answer_text} ({answer_link})"

    def parse_result(self, res):

        tool_call_font = "<font color='dark red'>TOOL CALL:</font></b>"
        tool_res_font = "<font color='yellow'>TOOL RESULT:</font></b>"
        query_font = "<font color='green'><b> USER QUERY:</font></b>"
        response_font = "<font color='green'><b>AGENT RESPONSE:</font></b>"

        self.printmd(f"{query_font} {res.text}")

        for action in res.generative_info.action_tracing_info.actions:

            if "tool_use" in action:
                tool_name = action.tool_use.action
                input_params = {}
                output_params = {}

                input_param = action.tool_use.input_action_parameters
                output_param = action.tool_use.output_action_parameters

                if isinstance(input_param, maps.MapComposite):
                    processed_input_params = self.recurse_proto_marshal_to_dict(
                        input_param
                    )
                    input_keys = list(processed_input_params.keys())
                    first_key = input_keys[0] if input_keys else None
                    input_params = processed_input_params.get(first_key, {})

                if isinstance(output_param, maps.MapComposite):
                    processed_output_params = (
                        self.recurse_proto_marshal_to_dict(output_param)
                    )
                    output_keys = list(processed_output_params.keys())
                    first_key = output_keys[0] if output_keys else None

                    output_params = processed_output_params.get(first_key, {})

                self.printmd(f"{tool_call_font} {tool_name} -> {input_params}")
                self.printmd(f"{tool_res_font} {output_params}")

            elif "agent_utterance" in action:
                self.printmd(f"{response_font} {action.agent_utterance.text}")



"""Unit Tests for Playbooks."""

# pylint: disable=redefined-outer-name
# pylint: disable=protected-access


# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from unittest.mock import patch, MagicMock
from dfcx_scrapi.core.playbooks import Playbooks
from google.cloud.dialogflowcx_v3beta1 import types
from google.cloud.dialogflowcx_v3beta1 import services
from google.protobuf import field_mask_pb2

@pytest.fixture
def test_config():
    project_id = "my-project-id-1234"
    location_id = "global"
    parent = f"projects/{project_id}/locations/{location_id}"
    agent_id = f"{parent}/agents/my-agent-1234"
    playbook_id = f"{agent_id}/playbooks/1234"
    goal = """You are a Google caliber software engineer that helps users write
        code."""
    instructions_list = [
        "Help the users write code snippets in python.",
        "Use ${TOOL: PLACEHOLDER} to help write code!"
        ]
    instructions_str = """
- Step 1
  - Step 1.1
- Step 2
  - Step 2.1
    - Step 2.1.1
    - Step 2.1.2
     - Step 2.1.2.1
- Step 3
"""
    instructions_proto_from_list = types.Playbook.Instruction(
        steps=[
        types.Playbook.Step(
            text="Help the users write code snippets in python."
            ),
        types.Playbook.Step(
            text="Use ${TOOL: PLACEHOLDER} to help write code!"
            )
            ]
        )

    # Note that we don't want any leading `-` in the final proto text because
    # the UI / console automatically adds this in. If you include the `-` then
    # you will end up with double leading `- -` in the console.
    instructions_proto_from_str = types.Playbook.Instruction(
        steps=[
            types.Playbook.Step(
                text="Step 1",
                steps=[
                    types.Playbook.Step(
                        text="Step 1.1"
                    )
                ]
            ),
            types.Playbook.Step(
                text="Step 2",
                steps=[
                    types.Playbook.Step(
                        text="Step 2.1",
                        steps=[
                            types.Playbook.Step(
                                text="Step 2.1.1"
                            ),
                            types.Playbook.Step(
                                text="Step 2.1.2",
                                steps=[
                                    types.Playbook.Step(
                                        text="Step 2.1.2.1"
                                    )
                                ]
                            )
                        ]
                    )
                ]
            ),
            types.Playbook.Step(
                text="Step 3"
            )
        ]
    )

    playbook_version_description = "v1.0"

    return {
        "project_id": project_id,
        "agent_id": agent_id,
        "playbook_id": playbook_id,
        "goal": goal,
        "instructions_list": instructions_list,
        "instructions_str": instructions_str,
        "instructions_proto_from_list": instructions_proto_from_list,
        "instructions_proto_from_str": instructions_proto_from_str,
        "playbook_version_description": playbook_version_description
    }

@pytest.fixture
def mock_playbook_obj_empty_instructions(test_config):
    return types.Playbook(
        name=test_config["playbook_id"],
        display_name="mock playbook",
        goal=test_config["goal"]
    )

@pytest.fixture
def mock_playbook_obj_list(test_config):
    return types.Playbook(
        name=test_config["playbook_id"],
        display_name="mock playbook",
        goal=test_config["goal"],
        instruction=test_config["instructions_proto_from_list"]
    )

@pytest.fixture
def mock_playbook_obj_str(test_config):
    return types.Playbook(
        name=test_config["playbook_id"],
        display_name="mock playbook",
        goal=test_config["goal"],
        instruction=test_config["instructions_proto_from_str"]
    )

@pytest.fixture
def mock_playbook_version_obj_no_description(
    test_config, mock_playbook_obj_empty_instructions):
    return types.PlaybookVersion(
        name=test_config["playbook_id"],
        playbook=mock_playbook_obj_empty_instructions,
    )

@pytest.fixture
def mock_playbook_version_obj_with_description(
    test_config, mock_playbook_obj_empty_instructions):
    return types.PlaybookVersion(
        name=test_config["playbook_id"],
        description=test_config["playbook_version_description"],
        playbook=mock_playbook_obj_empty_instructions,
    )


@pytest.fixture
def mock_agent_obj(test_config):
    return types.Agent(
        name=test_config["agent_id"],
        display_name="mock agent",
        default_language_code="en",
        time_zone="America/Chicago"
    )

@pytest.fixture
def mock_updated_playbook_obj(mock_playbook_obj_list):
    mock_playbook_obj_list.display_name = "mock playbook updated"
    return mock_playbook_obj_list


@pytest.fixture
def mock_list_playbooks_pager(mock_playbook_obj_list):
    return services.playbooks.pagers.ListPlaybooksPager(
        services.playbooks.PlaybooksClient.list_playbooks,
        types.playbook.ListPlaybooksRequest(),
        types.playbook.ListPlaybooksResponse(
            playbooks=[mock_playbook_obj_list]),
    )


@pytest.fixture(autouse=True)
def mock_client(test_config):
    """Fixture to create a mocked PlaybooksClient."""
    with patch("dfcx_scrapi.core.scrapi_base.default") as mock_default, \
        patch("dfcx_scrapi.core.scrapi_base.Request") as mock_request, \
        patch("dfcx_scrapi.core.playbooks.services.playbooks.PlaybooksClient") as mock_client, \
        patch("dfcx_scrapi.core.agents.Agents.__init__") as mock_agents_init:

        mock_creds = MagicMock()
        mock_default.return_value = (mock_creds, test_config["project_id"])
        mock_request.return_value = MagicMock()
        mock_agents_init.return_value = None

        yield mock_client 

@pytest.fixture
def mock_agents_client(test_config):
    """Fixture to create a mocked AgentsClient."""
    with patch("dfcx_scrapi.core.scrapi_base.default") as mock_default, \
        patch("dfcx_scrapi.core.scrapi_base.Request") as mock_request, \
        patch("dfcx_scrapi.core.agents.services.agents.AgentsClient") as mock_client:
        mock_creds = MagicMock()
        mock_default.return_value = (mock_creds, test_config["project_id"])
        mock_request.return_value = MagicMock()

        yield mock_client # Return control to test method

# Test get_playbooks_map
def test_get_playbooks_map(mock_client, mock_list_playbooks_pager, test_config):
    pb = Playbooks(agent_id=test_config["agent_id"])
    mock_client.return_value.list_playbooks.return_value = mock_list_playbooks_pager # pylint: disable=C0301
    res = pb.get_playbooks_map(agent_id=test_config["agent_id"])

    assert isinstance(res, dict)
    assert test_config["playbook_id"] in res
    assert res[test_config["playbook_id"]] == "mock playbook"


# Test list_playbooks
def test_list_playbooks(mock_client, mock_list_playbooks_pager, test_config):
    pb = Playbooks(agent_id=test_config["agent_id"])
    mock_client.return_value.list_playbooks.return_value = mock_list_playbooks_pager # pylint: disable=C0301
    res = pb.list_playbooks()

    assert isinstance(res, list)
    assert isinstance(res[0], types.Playbook)


# Test get_playbook
def test_get_playbook(mock_client, mock_playbook_obj_list, test_config):
    pb = Playbooks(agent_id=test_config["agent_id"])
    mock_client.return_value.get_playbook.return_value = mock_playbook_obj_list
    res = pb.get_playbook(playbook_id=test_config["playbook_id"])

    assert isinstance(res, types.Playbook)
    assert res.display_name == "mock playbook"


# Test create_playbook
def test_create_playbook_from_kwargs_instruction_list(
        mock_client, mock_playbook_obj_list, test_config):
    pb = Playbooks(agent_id=test_config["agent_id"])
    mock_client.return_value.create_playbook.return_value = mock_playbook_obj_list # pylint: disable=C0301
    res = pb.create_playbook(
        agent_id=test_config["agent_id"],
        display_name="mock playbook",
        goal=test_config["goal"],
        instructions=test_config["instructions_list"]
    )
    assert isinstance(res, types.Playbook)
    assert res.display_name == "mock playbook"
    assert res.instruction == test_config["instructions_proto_from_list"]

def test_create_playbook_from_kwargs_instruction_str(
    mock_client, mock_playbook_obj_str, test_config):
    pb = Playbooks(agent_id=test_config["agent_id"])
    mock_client.return_value.create_playbook.return_value = mock_playbook_obj_str # pylint: disable=C0301
    res = pb.create_playbook(
        agent_id=test_config["agent_id"],
        display_name="mock playbook",
        goal=test_config["goal"],
        instructions=test_config["instructions_str"]
    )
    assert isinstance(res, types.Playbook)
    assert res.display_name == "mock playbook"
    assert res.instruction == test_config["instructions_proto_from_str"]

def test_create_playbook_from_proto_object(
    mock_client, mock_playbook_obj_list, test_config):
    pb = Playbooks(agent_id=test_config["agent_id"])
    mock_client.return_value.create_playbook.return_value = mock_playbook_obj_list # pylint: disable=C0301
    res = pb.create_playbook(
        agent_id=test_config["agent_id"],
        obj=mock_playbook_obj_list
    )
    assert isinstance(res, types.Playbook)
    assert res.display_name == "mock playbook"


# Test update_playbook
def test_update_playbook_with_obj(
    mock_client, mock_updated_playbook_obj, test_config):
    pb = Playbooks(agent_id=test_config["agent_id"])
    mock_client.return_value.update_playbook.return_value = (
        mock_updated_playbook_obj
    )
    res = pb.update_playbook(
        playbook_id=test_config["playbook_id"],
        obj=mock_updated_playbook_obj
    )

    assert isinstance(res, types.Playbook)
    assert res.display_name == "mock playbook updated"


def test_update_playbook_with_kwargs(
    mock_client, mock_playbook_obj_list, test_config):
    pb = Playbooks(agent_id=test_config["agent_id"])
    mock_client.return_value.get_playbook.return_value = mock_playbook_obj_list
    mock_client.return_value.update_playbook.return_value = mock_playbook_obj_list # pylint: disable=C0301
    res = pb.update_playbook(
        playbook_id=test_config["playbook_id"],
        display_name="mock playbook updated"
    )

    assert isinstance(res, types.Playbook)
    assert res.display_name == "mock playbook updated"

# Test the playbook kwarg processing helper methods
def test_process_playbook_kwargs_display_name(
        mock_playbook_obj_str, mock_updated_playbook_obj, test_config):
    pb = Playbooks(agent_id=test_config["agent_id"])
    kwargs = {"display_name": "mock playbook updated"}

    expected_mask = field_mask_pb2.FieldMask(paths=["display_name"])
    playbook, mask = pb.process_playbook_kwargs(mock_playbook_obj_str, **kwargs)

    assert mock_updated_playbook_obj.display_name == playbook.display_name
    assert expected_mask == mask

def test_process_playbook_kwargs_instruction_list(
        mock_playbook_obj_empty_instructions,
        mock_playbook_obj_list, test_config):
    pb = Playbooks(agent_id=test_config["agent_id"])

    # patch the object so we can track the internal method call
    with patch.object(
        pb, "build_instructions_from_list",
        wraps=pb.build_instructions_from_list) as mock_build_instructions:

        kwargs = {"instructions": test_config["instructions_list"]}
        expected_mask = field_mask_pb2.FieldMask(paths=["instruction"])

        playbook, mask = pb.process_playbook_kwargs(
            mock_playbook_obj_empty_instructions, **kwargs)

        assert mock_playbook_obj_list.instruction == playbook.instruction
        assert expected_mask == mask
        mock_build_instructions.assert_called_once_with(
            test_config["instructions_list"])

def test_process_playbook_kwargs_instruction_str(
        mock_playbook_obj_empty_instructions,
        mock_playbook_obj_str, test_config):
    pb = Playbooks(agent_id=test_config["agent_id"])

    # patch the object so we can track the internal method call
    with patch.object(
        pb, "build_instructions_from_string",
        wraps=pb.build_instructions_from_string) as mock_build_instructions:

        kwargs = {"instructions": test_config["instructions_str"]}
        expected_mask = field_mask_pb2.FieldMask(paths=["instruction"])

        playbook, mask = pb.process_playbook_kwargs(
            mock_playbook_obj_empty_instructions, **kwargs)

        assert mock_playbook_obj_str.instruction == playbook.instruction
        assert expected_mask == mask
        mock_build_instructions.assert_called_once_with(
            test_config["instructions_str"]
        )

def test_process_playbook_kwargs_instruction_obj(
        mock_playbook_obj_empty_instructions,
        mock_playbook_obj_str, test_config):
    pb = Playbooks(agent_id=test_config["agent_id"])
    kwargs = {"instructions": test_config["instructions_proto_from_str"]}
    expected_mask = field_mask_pb2.FieldMask(paths=["instruction"])

    playbook, mask = pb.process_playbook_kwargs(
        mock_playbook_obj_empty_instructions, **kwargs)

    assert mock_playbook_obj_str.instruction == playbook.instruction
    assert expected_mask == mask

# Test delete_playbook
def test_delete_playbook(mock_client, test_config):
    pb = Playbooks(agent_id=test_config["agent_id"])
    pb.delete_playbook(playbook_id=test_config["playbook_id"])
    mock_client.return_value.delete_playbook.assert_called()

# Test set_default_playbook
def test_set_default_playbook(mock_agents_client, mock_agent_obj, test_config):
    pb = Playbooks(agent_id=test_config["agent_id"])
    mock_agents_client.return_value.get_agent.return_value = mock_agent_obj
    mock_agents_client.return_value.update_agent.return_value = mock_agent_obj
    pb.set_default_playbook(playbook_id=test_config["playbook_id"])

    assert mock_agent_obj.start_playbook == test_config["playbook_id"]

# Test build instruction helpers
def test_build_instructions_from_list(test_config):
    pb = Playbooks(agent_id=test_config["agent_id"])
    res = pb.build_instructions_from_list(
        instructions=test_config["instructions_list"])

    assert res == test_config["instructions_proto_from_list"]

def test_build_instructions_from_str(test_config):
    pb = Playbooks(agent_id=test_config["agent_id"])
    res = pb.build_instructions_from_string(
        instructions=test_config["instructions_str"])

    assert res == test_config["instructions_proto_from_str"]

def test_parse_steps_simple_list(test_config):
    pb = Playbooks(agent_id=test_config["agent_id"])

    lines = [
        "Step 1",
        "Step 2",
        "Step 3"
    ]

    expected_steps = [
        types.Playbook.Step(text="Step 1"),
        types.Playbook.Step(text="Step 2"),
        types.Playbook.Step(text="Step 3")
    ]

    steps, next_index = pb.parse_steps(lines, 0, 0)
    assert steps == expected_steps
    assert next_index == 3

def test_parse_steps_nested_list(test_config):
    pb = Playbooks(agent_id=test_config["agent_id"])

    lines = [
        "- Step 1",
        "  - Step 1.1",
        "- Step 2",
        "  - Step 2.1",
        "    - Step 2.1.1",
        "    - Step 2.1.2",
        "     - Step 2.1.2.1",
        "- Step 3"
        ]

    steps, next_index = pb.parse_steps(lines, 0, 0)
    assert steps == test_config["instructions_proto_from_str"].steps
    assert next_index == 8

def test_create_playbook_version_no_description(
        mock_client, test_config, mock_playbook_version_obj_no_description):
    pb = Playbooks(agent_id=test_config["agent_id"])

    mock_client.return_value.create_playbook_version.return_value = mock_playbook_version_obj_no_description

    res = pb.create_playbook_version(playbook_id=test_config["playbook_id"])

    mock_client.return_value.create_playbook_version.assert_called()
    assert isinstance(res, types.PlaybookVersion)
    assert res.playbook.name == test_config["playbook_id"]
    assert res.description == ""

def test_create_playbook_version_with_description(
        mock_client, test_config, mock_playbook_version_obj_with_description):
    pb = Playbooks(agent_id=test_config["agent_id"])

    mock_client.return_value.create_playbook_version.return_value = mock_playbook_version_obj_with_description

    res = pb.create_playbook_version(playbook_id=test_config["playbook_id"])

    mock_client.return_value.create_playbook_version.assert_called()
    assert isinstance(res, types.PlaybookVersion)
    assert res.playbook.name == test_config["playbook_id"]
    assert res.description == test_config["playbook_version_description"]



"""Test Class for core Flow Methods in SCRAPI."""

# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from typing import Dict
from unittest.mock import patch, MagicMock

from google.cloud.dialogflowcx_v3beta1 import types
from google.cloud.dialogflowcx_v3beta1.services.flows import pagers, FlowsClient

from dfcx_scrapi.core.flows import Flows

@pytest.fixture
def test_config():
    project_id = "my-project-id-1234"
    location_id = "global"
    parent = f"projects/{project_id}/locations/{location_id}"
    agent_id = f"{parent}/agents/my-agent-1234"
    display_name = "My Agent Display Name"
    default_id = "00000000-0000-0000-0000-000000000000"
    another_id = "99999999-9999-9999-9999-999999999999"
    start_flow_id = f"{agent_id}/flows/{default_id}"
    another_flow_id = f"{agent_id}/flows/{another_id}"

    return {
        "project_id": project_id,
        "agent_id": agent_id,
        "display_name": display_name,
        "location_id": location_id,
        "start_flow_id": start_flow_id,
        "another_flow_id": another_flow_id
    }

@pytest.fixture
def nlu_settings():
    return types.NluSettings(
        model_type=types.NluSettings.ModelType.MODEL_TYPE_ADVANCED,
        classification_threshold=0.3,
        model_training_mode=types.NluSettings.ModelTrainingMode.MODEL_TRAINING_MODE_AUTOMATIC
    )

@pytest.fixture
def mock_flows_list(test_config, nlu_settings):
    flow1 = types.Flow(
        name = test_config["start_flow_id"],
        display_name = "Default Start Flow",
        description = "This is the default start flow.",
        nlu_settings = nlu_settings
    )

    flow2 = types.Flow(
        name = test_config["another_flow_id"],
        display_name = "Flow #2",
        description = "Another Flow",
        nlu_settings = nlu_settings
    )

    return [flow1, flow2]

@pytest.fixture
def mock_list_flows_pager(mock_flows_list):
    return pagers.ListFlowsPager(
        FlowsClient.list_flows,
        types.ListFlowsRequest(),
        types.ListFlowsResponse(
            flows=mock_flows_list),
    )

@pytest.fixture(autouse=True)
def mock_creds(test_config: Dict[str, str]):
    """Setup fixture for Flows Class to be used with all tests."""
    with patch("dfcx_scrapi.core.scrapi_base.default") as mock_default, \
        patch("dfcx_scrapi.core.scrapi_base.Request") as mock_request:
        mock_creds = MagicMock()
        mock_default.return_value = (mock_creds, test_config["project_id"])
        mock_request.return_value = MagicMock()

        yield mock_creds

def test_build_nlu_settings_standard():
   nlu_settings = Flows._build_nlu_settings()
   assert nlu_settings.model_type == types.NluSettings.ModelType.MODEL_TYPE_STANDARD
   assert round(nlu_settings.classification_threshold, 1) == 0.3
   assert nlu_settings.model_training_mode == types.NluSettings \
    .ModelTrainingMode.MODEL_TRAINING_MODE_MANUAL

def test_build_nlu_settings_advanced():
    nlu_settings = Flows._build_nlu_settings(
       model_type="ADVANCED",
       classification_threshold=0.7,
       model_training_mode="AUTOMATIC"
       )

    assert nlu_settings.model_type == types.NluSettings.ModelType.MODEL_TYPE_ADVANCED
    assert round(nlu_settings.classification_threshold,1) == 0.7
    assert nlu_settings.model_training_mode == types.NluSettings \
        .ModelTrainingMode.MODEL_TRAINING_MODE_AUTOMATIC

def test_build_nlu_settings_invalid_model_type():
    with pytest.raises(KeyError):
       Flows._build_nlu_settings(model_type="INVALID")


def test_build_nlu_settings_invalid_training_mode():
    with pytest.raises(KeyError):
        Flows._build_nlu_settings(model_training_mode="INVALID")

@patch("dfcx_scrapi.core.flows.services.flows.FlowsClient.list_flows")
def test_get_flows_map(mock_list_flows, mock_list_flows_pager, test_config):
    mock_list_flows.return_value = mock_list_flows_pager

    flow = Flows()
    flows_map = flow.get_flows_map(agent_id=test_config["agent_id"])

    assert flows_map == {
        test_config["start_flow_id"]: "Default Start Flow",
        test_config["another_flow_id"]: "Flow #2"
        }



"""A set of builder methods to create CX proto resource objects"""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import List

from google.cloud.dialogflowcx_v3beta1.types.generative_settings import GenerativeSettings
from google.cloud.dialogflowcx_v3beta1.types.safety_settings import SafetySettings
from dfcx_scrapi.builders.builders_common import BuildersCommon

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class GenerativeSettingsBuilder(BuildersCommon):
    """Base Class for Generative Settings builder."""

    _proto_type = GenerativeSettings
    _proto_type_str = "GenerativeSettings"

    def __str__(self) -> str:
        """String representation of the proto_obj."""
        self._check_proto_obj_attr_exist()

        return (
            f"\nlanguage_code: {self.proto_obj.language_code}"
            f"\nfallback_settings: {self.proto_obj.fallback_settings}"
            f"\ngenerative_safety_settings: "
                f"{self.proto_obj.generative_safety_settings}"
            f"\nknowledge_connector_settings: "
                f"{self.proto_obj.knowledge_connector_settings}"
            )

    def fallback_settings(
            self, selected_prompt: str,
            prompt_templates: List[
                GenerativeSettings.FallbackSettings.PromptTemplate]
                ) -> GenerativeSettings.FallbackSettings:
        """Build the Fallback Settings proto."""
        return GenerativeSettings.FallbackSettings(
            selected_prompt=selected_prompt,
            prompt_templates=prompt_templates
        )

    def prompt_template(
            self, display_name: str, prompt_text: str, frozen: bool = False
    ):
        """Build the Prompt Template proto."""
        return GenerativeSettings.FallbackSettings.PromptTemplate(
            display_name=display_name,
            prompt_text=prompt_text,
            frozen=frozen
        )

    def banned_phrases(self, phrases: List[str],
                       language_code: str = "en") -> SafetySettings:
        """Build the SafetySettings proto with banned phrases."""
        banned_phrases = []
        for phrase in phrases:
            phrase_proto = SafetySettings.Phrase(
                text=phrase,
                language_code=language_code
            )
            banned_phrases.append(phrase_proto)

        safety_settings = SafetySettings(
            banned_phrases=banned_phrases
        )

        return safety_settings

    def knowledge_connector_settings(
            self, business_name: str, agent_name: str = None,
            agent_identity: str = None, business_description: str = None,
            agent_scope: str = None
            ) -> GenerativeSettings.KnowledgeConnectorSettings:
        """Build the knowledge settings proto.

        Args:
          business_name: The name of the company or business that corresponds
            most closely with the documents in the Data Store(s).
          agent_name: The name or identify of the conversational Agent. Ex:
            Bard, Fred, Julia, MedBot, etc.
          agent_identity: General description of the type of conversational
            Agent that this is. Ex: AI Assistant, Travel Concierge, Medical
            Assistant, etc.
          business_description: A brief description of the business and what it
            does
          agent_scope: The defined scope of this conversational Agent and how
            it should (or should not) interact with users.
        """
        return GenerativeSettings.KnowledgeConnectorSettings(
            business=business_name,
            agent=agent_name,
            agent_identity=agent_identity,
            business_description=business_description,
            agent_scope=agent_scope
        )



"""Examples class for Generative Agents."""

# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dfcx_scrapi.core import scrapi_base
from dfcx_scrapi.core import playbooks
from dfcx_scrapi.core import tools
from typing import Dict, List, Any

from google.cloud.dialogflowcx_v3beta1 import services
from google.cloud.dialogflowcx_v3beta1 import types
from google.protobuf import field_mask_pb2


class Examples(scrapi_base.ScrapiBase):
    """Core Class for CX Examples Resource functions."""

    def __init__(
        self,
        agent_id: str,
        creds_path: str = None,
        creds_dict: Dict = None,
        creds=None,
        scope=False,
        example_id: str = None,
        playbook_id: str = None,
        playbooks_map: Dict[str, str] = None,
        tools_map: Dict[str, str] = None
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            scope=scope,
        )

        self.agent_id = agent_id

        client_options = self._set_region(self.agent_id)
        self.examples_client = services.examples.ExamplesClient(
            credentials=self.creds, client_options=client_options
        )
        self.playbooks_client = playbooks.Playbooks(agent_id=self.agent_id)
        self.tools_client = tools.Tools(agent_id=self.agent_id)

        self.playbook_id = playbook_id
        self.example_id = example_id
        self.playbooks_map = playbooks_map
        self.tools_map = tools_map

    @staticmethod
    def get_playbook_state(playbook_state: str):
        """Simple converter for enum values on playbook state."""
        if playbook_state == "OK":
            return 1
        elif playbook_state == "CANCELLED":
            return 2
        elif playbook_state == "FAILED":
            return 3
        elif playbook_state == "ESCALATED":
            return 4
        elif playbook_state == "PENDING":
            return 5
        else:
            return 0

    def build_example_from_action_list(
        self,
        display_name: str,
        action_list: List[Dict],
        description: str = None
        ) -> types.Example:
        """Builds an Example from a list of action dictionaries."""

        example = types.Example()
        example.display_name = display_name
        example.description = description

        for action_dict in action_list:
            action_type, action_data = next(iter(action_dict.items()))

            if action_type == "user_utterance":
                example.actions.append(
                    types.Action(
                        user_utterance=types.UserUtterance(text=action_data)))

            elif action_type == "agent_utterance":
                example.actions.append(
                    types.Action(
                        agent_utterance=types.AgentUtterance(text=action_data)))

            elif action_type == "tool_use":
                if not self.tools_map:
                    self.tools_map = self.tools_client.get_tools_map(
                        self.agent_id, reverse=True
                    )

                tool_name = action_data.get("tool_name")

                action = types.Action()
                tool_use = types.ToolUse()
                tool_use.tool = self.tools_map.get(tool_name)
                tool_use.action = action_data.get("action")
                tool_use.input_action_parameters = action_data.get(
                    "input_action_parameters", None)
                tool_use.output_action_parameters = action_data.get(
                    "output_action_parameters", None)

                action.tool_use = tool_use
                example.actions.append(action)

            elif action_type == "playbook_invocation":
                example.actions.append(
                    types.Action(
                        playbook_invocation=self.build_playbook_invocation(
                            action_data)))

        return example

    def build_playbook_invocation(
        self, action: Dict[str, Any]) -> types.PlaybookInvocation:
        """Helper method for constructing Playbook invocation."""
        if not self.playbooks_map:
            self.playbooks_map = self.playbooks_client.get_playbooks_map(
                self.agent_id, reverse=True
            )

        pb_inv = types.PlaybookInvocation()
        pb_inv.playbook = self.playbooks_map[action.get("playbook_name", None)]

        pb_input = types.PlaybookInput()
        pb_input.preceding_conversation_summary = action.get(
            "playbook_input_summary", None)
        input_params = action.get("playbook_input_parameters", None)
        if input_params:
            pb_input.action_parameters = input_params

        pb_output = types.PlaybookOutput()
        pb_output.execution_summary = action.get(
            "playbook_output_summary", None)
        output_params = action.get("playbook_output_parameters", None)
        if output_params:
            pb_output.action_parameters = output_params

        pb_state = types.OutputState(
            self.get_playbook_state(action.get("playbook_state", None)))

        pb_inv.playbook_input = pb_input
        pb_inv.playbook_output = pb_output
        pb_inv.playbook_state = pb_state

        return pb_inv

    def get_examples_map(self, playbook_id: str = None, reverse=False):
        """Exports Agent Example Names and UUIDs into a user friendly dict.

        Args:
          playbook_id: the formatted CX Agent Playbook ID to use
          reverse: (Optional) Boolean flag to swap key:value -> value:key

        Returns:
          Dict containing Example UUIDs as keys and display names as values
        """
        if not playbook_id:
            playbook_id = self.playbook_id

        if reverse:
            examples_dict = {
                example.display_name: example.name
                for example in self.list_examples(playbook_id)
            }

        else:
            examples_dict = {
                example.name: example.display_name
                for example in self.list_examples(playbook_id)
            }

        return examples_dict

    @scrapi_base.api_call_counter_decorator
    def list_examples(
        self,
        playbook_id: str = None,
        language_code: str = "en") -> List[types.Example]:
        """Get a List of all Examples in the specified Playbook.

        Args:
          playbook_id: the properly formatted Playbook ID string
          language_code: Specifies the language of the Examples listed.

        Returns:
          A List of CX Example objects for the specific Playbook ID
        """
        if not playbook_id:
            playbook_id = self.playbook_id

        request = types.example.ListExamplesRequest()
        request.parent = playbook_id
        request.language_code = language_code

        client_options = self._set_region(playbook_id)
        client = services.examples.ExamplesClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.list_examples(request)

        cx_examples = []
        for page in response.pages:
            for cx_example in page.examples:
                cx_examples.append(cx_example)

        return cx_examples

    @scrapi_base.api_call_counter_decorator
    def get_example(self, example_id: str = None) -> types.Example:
        """Get a single CX Example object based on the provided Example ID.

        Args:
          example_id: a properly formatted CX Example ID

        Returns:
          A single CX Example Object
        """
        if not example_id:
            example_id = self.example_id

        client_options = self._set_region(example_id)
        client = services.examples.ExamplesClient(
            credentials=self.creds, client_options=client_options
        )

        response = client.get_example(name=example_id)

        return response

    @scrapi_base.api_call_counter_decorator
    def create_example(
        self,
        playbook_id: str = None,
        obj: types.Example = None,
        **kwargs
        ) -> types.Example:
        """Create a single CX Example object in the specified Playbook ID.

        Args:
          playbook_id: the CX Playbook ID where the Example will be created
          obj: (Optional) a CX Example object of types.Example

        Returns:
          A copy of the successful Example object that was created
        """
        if not playbook_id:
            playbook_id = self.playbook_id

        if obj:
            example = obj
            example.name = ""
        else:
            example = types.example.Example()

        for key, value in kwargs.items():
            setattr(example, key, value)

        client_options = self._set_region(playbook_id)
        client = services.examples.ExamplesClient(
            credentials=self.creds, client_options=client_options
        )

        response = client.create_example(parent=playbook_id, example=example)

        return response

    @scrapi_base.api_call_counter_decorator
    def update_example(
        self,
        example_id: str = None,
        obj: types.Example = None,
        **kwargs
        ) -> types.Example:
        """Update a single CX Example object.

        Args:
          example_id: the CX Example ID to update
          obj: (Optional) a CX Example object of types.Example

        Returns:
          A copy of the successful Example object that was created
        """
        if obj:
            example = obj
            example.name = example_id
        else:
            if not example_id:
                example_id = self.example_id
            example = self.get_example(example_id)

        for key, value in kwargs.items():
            setattr(example, key, value)
        paths = kwargs.keys()
        mask = field_mask_pb2.FieldMask(paths=paths)

        client_options = self._set_region(example_id)
        client = services.examples.ExamplesClient(
            credentials=self.creds, client_options=client_options
        )

        response = client.update_example(example=example, update_mask=mask)

        return response

    @scrapi_base.api_call_counter_decorator
    def delete_example(
        self, example_id: str = None,
        obj: types.Example = None):
        """Deletes the specified Example.

        Args:
          example_id: The formatted CX Example ID to delete.
          obj: (Optional) a CX Example object of types.Example
        """
        if not example_id:
            example_id = self.example_id

        if obj:
            example_id = obj.name

        client_options = self._set_region(example_id)
        client = services.examples.ExamplesClient(
            credentials=self.creds, client_options=client_options)
        client.delete_example(name=example_id)



"""Page Resource functions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Dict, List
from google.cloud.dialogflowcx_v3beta1.services import pages
from google.cloud.dialogflowcx_v3beta1.types import page as gcdc_page
from google.protobuf import field_mask_pb2

from dfcx_scrapi.core import scrapi_base

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class Pages(scrapi_base.ScrapiBase):
    """Core Class for CX Page Resource functions."""

    def __init__(
        self,
        creds_path: str = None,
        creds_dict: Dict = None,
        scope=None,
        creds=None,
        page_id: str = None,
        flow_id: str = None,
        language_code: str = "en"
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            scope=scope,
        )

        self.page_id = page_id
        self.flow_id = flow_id
        self.language_code = language_code

    @staticmethod
    def _add_generic_pages_to_map(flow_id, pages_map, reverse):
        """Add the generic page names to each Page map.

        Dialogflow CX contains a few `special` pages names that are reserved
        and do not have UUID4 format IDs. This will take the existing page
        map and insert them in for downstream lookups.

        Args:
          flow_id: The Flow ID that contains the Pages in the pages_map
          pages_map: The existing pages_map Dict from `get_pages_map` that
            we will add the new special pages to.
          reverse: Boolean flag to swap key:value -> value:key
        """
        page_names = ["START_PAGE", "END_FLOW", "END_SESSION"]

        if reverse:
            for page in page_names:
                pages_map[page] = f"{flow_id}/pages/{page}"
        else:
            for page in page_names:
                pages_map[f"{flow_id}/pages/{page}"] = page

        return pages_map

    def get_pages_map(
        self, flow_id: str = None, reverse=False
    ) -> Dict[str, str]:
        """Exports Agent Page UUIDs and Names into a user friendly dict.

        Args:
          flow_id: the formatted CX Agent Flow ID to use
          reverse: (Optional) Boolean flag to swap key:value -> value:key

        Returns:
          Dictionary containing Page UUIDs as keys and display names as values.
          If Optional reverse=True, the output will return page_name:ID mapping
          instead of ID:page_name
        """
        if not flow_id:
            flow_id = self.flow_id

        if reverse:
            pages_dict = {
                page.display_name: page.name
                for page in self.list_pages(flow_id)
            }

        else:
            pages_dict = {
                page.name: page.display_name
                for page in self.list_pages(flow_id)
            }

        pages_dict = self._add_generic_pages_to_map(
            flow_id, pages_dict, reverse
        )

        return pages_dict

    @scrapi_base.api_call_counter_decorator
    def list_pages(
        self,
        flow_id: str = None,
        language_code: str = None) -> List[gcdc_page.Page]:
        """Get a List of all pages for the specified Flow ID.

        Args:
          flow_id: the properly formatted Flow ID string
          language_code: Specifies the language of the Pages listed. While the
            majority of contents of a Page is language agnostic, the contents
            in the "Agent Says" and similar parts of a Page are affected by
            language code.

        Returns:
          A List of CX Page objects for the specific Flow ID
        """
        request = gcdc_page.ListPagesRequest()
        request.parent = flow_id

        # prefer method input over class level input
        if language_code:
            request.language_code = language_code
        else:
            request.language_code = self.language_code

        client_options = self._set_region(flow_id)
        client = pages.PagesClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.list_pages(request)

        cx_pages = []
        for page in response.pages:
            for cx_page in page.pages:
                cx_pages.append(cx_page)

        return cx_pages

    @scrapi_base.api_call_counter_decorator
    def get_page(self, page_id: str = None) -> gcdc_page.Page:
        """Get a single CX Page object based on the provided Page ID.

        Args:
          page_id: a properly formatted CX Page ID

        Returns:
          A single CX Page Object
        """
        if not page_id:
            page_id = self.page_id

        client_options = self._set_region(page_id)
        client = pages.PagesClient(
            credentials=self.creds, client_options=client_options
        )

        response = client.get_page(name=page_id)

        return response

    @scrapi_base.api_call_counter_decorator
    def create_page(
        self, flow_id: str = None, obj: gcdc_page.Page = None, **kwargs
    ) -> gcdc_page.Page:
        """Create a single CX Page object in the specified Flow ID.

        Args:
          flow_id: the CX Flow ID where the Page object will be created
          obj: (Optional) a CX Page object of gcdc_page.Page

        Returns:
          A copy of the successful Page object that was created
        """
        if not flow_id:
            flow_id = self.flow_id

        if obj:
            page = obj
            page.name = ""
        else:
            page = gcdc_page.Page()

        for key, value in kwargs.items():
            setattr(page, key, value)

        client_options = self._set_region(flow_id)
        client = pages.PagesClient(
            credentials=self.creds, client_options=client_options
        )

        response = client.create_page(parent=flow_id, page=page)

        return response

    @scrapi_base.api_call_counter_decorator
    def update_page(
        self, page_id: str = None, obj: gcdc_page.Page = None, **kwargs
    ) -> gcdc_page.Page:
        """Update a single CX Page object.

        Args:
          page_id: the CX Page ID to update
          obj: (Optional) a CX Page object of gcdc_page.Page

        Returns:
          A copy of the successful Page object that was created
        """
        if obj:
            page = obj
            page.name = page_id
        else:
            if not page_id:
                page_id = self.page_id
            page = self.get_page(page_id)

        for key, value in kwargs.items():
            setattr(page, key, value)
        paths = kwargs.keys()
        mask = field_mask_pb2.FieldMask(paths=paths)

        client_options = self._set_region(page_id)
        client = pages.PagesClient(
            credentials=self.creds, client_options=client_options
        )

        response = client.update_page(page=page, update_mask=mask)

        return response

    @scrapi_base.api_call_counter_decorator
    def delete_page(
        self, page_id: str = None,
        obj: gcdc_page.Page = None, force: bool = False
    ):
        """Deletes the specified Page.

        Args:
          page_id: The formatted CX Page ID to delete.
          obj: (Optional) a CX Page object of gcdc_page.Page
          force: (Optional) This field has no effect for pages with no incoming
            transitions. If set to True, Dialogflow will remove the page,
            as well as any transitions to the page.
        """
        if not page_id:
            page_id = self.page_id

        if obj:
            page_id = obj.name

        client_options = self._set_region(page_id)
        client = pages.PagesClient(
            credentials=self.creds, client_options=client_options)
        req = gcdc_page.DeletePageRequest(name=page_id, force=force)
        client.delete_page(request=req)



"""Engines Resource functions for Vertex Search and Conversation."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Dict, List
from google.longrunning.operations_pb2 import Operation
from google.cloud.discoveryengine import (
    Engine,
    EngineServiceClient,
    ListEnginesRequest,
    GetEngineRequest,
    CreateEngineRequest,
    DeleteEngineRequest
    )

from dfcx_scrapi.core import scrapi_base

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class Engines(scrapi_base.ScrapiBase):
    """Engines Class for Data Stores in Vertex Search and Conversation."""
    def __init__(
        self,
        project_id: str,
        creds_path: str = None,
        creds_dict: Dict = None,
        creds=None,
        scope=False,
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            scope=scope,
        )

        self.project_id = project_id

    def __process_data_store_ids(self, data_store_ids: List[str]) -> List[str]:
        """Process the data store IDs to ensure they are the correct format."""
        if not isinstance(data_store_ids, List):
            raise TypeError(f"Expected a List of Strings, got {data_store_ids}")

        processed_ids = []
        for ds_id in data_store_ids:
            processed_ids.append(self._validate_data_store_id(ds_id))

        return processed_ids

    def get_engines_map(
            self, location: str = "global", reverse: bool = False
            ) -> Dict[str, str]:
        """Get a user friendly mapping for Engine Names and IDs."""
        if reverse:
            engine_dict = {
                engine.display_name: engine.name
                for engine in self.list_engines(location=location)
            }

        else:
            engine_dict = {
                engine.name: engine.display_name
                for engine in self.list_engines(location=location)
            }

        return engine_dict

    def build_chat_engine_proto(
            self, display_name: str, business_name: str,
            data_store_ids: List[str], existing_df_agent_id: str = None,
            language_code: str = "en", time_zone: str = "America/Chicago"):
        """Build the Chat Engine proto for creating a new Engine.

        Args:
          display_name: the human readable display name of the Chat Engine
          business_name: the name of the company or business that corresponds
            most closely with the documents in the Data Store(s).
          existing_df_agent_id: the Dialogflow Agent ID to associate with the
            newly created Engine. Only provide this if linking to an existing
            Dialogflow CX Agent.
          language_code: the default language of the agent as a language tag
          time_zone: the time zone of the agent from the time zone database

        Returns:
          The Engine object configured as a ChatEngine.
        """
        data_store_ids = self.__process_data_store_ids(data_store_ids)

        engine = Engine()

        ce_config = Engine.ChatEngineConfig()

        agent_config = Engine.ChatEngineConfig.AgentCreationConfig()
        agent_config.business = business_name
        agent_config.default_language_code = language_code
        agent_config.time_zone = time_zone
        ce_config.agent_creation_config = agent_config

        if existing_df_agent_id:
            ce_config.dialogflow_agent_to_link = existing_df_agent_id

        engine.display_name = display_name
        engine.chat_engine_config = ce_config
        engine.solution_type = self._get_solution_type("chat")
        engine.data_store_ids = data_store_ids

        return engine


    def build_search_engine_proto(self, display_name: str, data_store_ids: List[str],
                                  search_tier: str = "SEARCH_TIER_STANDARD"):
        """Build the Search Engine proto for creating a new Engine.

        Args:
          display_name: the human readable display name of the Search Engine
          data_store_ids: a list of Data Store IDs associated with this engine.
          search_tier: either SEARCH_TIER_STANDARD (default) or SEARCH_TIER_ENTERPRISE

        Returns:
          The Engine object configured as a SearchEngine.
        """
        data_store_ids = self.__process_data_store_ids(data_store_ids)

        engine = Engine()

        se_config = Engine.SearchEngineConfig()
        se_config.search_tier = search_tier

        engine.display_name = display_name
        engine.solution_type = self._get_solution_type("search")
        engine.data_store_ids = data_store_ids
        engine.search_engine_config = se_config

        return engine


    def list_engines(
            self, location: str = "global") -> List[Engine]:
        """List all Engines for a given project and location."""
        parent = self._build_data_store_parent(location)
        client = EngineServiceClient()
        request = ListEnginesRequest(parent=parent)
        page_result = client.list_engines(request=request)

        engines = []
        for response in page_result:
            engines.append(response)

        return engines

    def get_engine(self, engine_id: str) -> Engine:
        """Get a single Engine by specified ID."""
        client = EngineServiceClient()
        request = GetEngineRequest(name=engine_id)
        response = client.get_engine(request=request)

        return response

    def create_engine(self, engine: Engine,
                      location: str = "global") -> Engine:
        """Create a new Vertex Search Engine based on user inputs.

        Use the DataStores.build_chat_engine_proto to create the proper Engine
        payload to pass in for the `engine` arg.

        Args:
          location, the GCP region to create the Engine in
          engine, a proto object of type discoveryengine.Engine
            Note that at this time only "Chat" engines are supported
          solution_type, "chat" is the only value supported at this time.
        """
        parent = self._build_data_store_parent(location)
        client_options = self._client_options_discovery_engine(parent)
        client = EngineServiceClient(
            credentials=self.creds, client_options=client_options
        )

        request = CreateEngineRequest(
            parent=parent,
            engine=engine,
            engine_id=engine.display_name
        )

        operation = client.create_engine(request=request)
        print("Waiting for operation to complete...")
        response = operation.result()
        print(f"Successfully created Engine: {response.display_name}")

        return response

    def delete_engine(self, engine_id: str) -> Operation:
        """Deletes the specified Engine."""
        client_options = self._client_options_discovery_engine(engine_id)
        client = EngineServiceClient(
            credentials=self.creds, client_options=client_options
        )

        request = DeleteEngineRequest(
            name=engine_id
        )
        operation = client.delete_engine(request=request)

        return operation.operation



"""Site Search Service functions for Vertex Search and Conversation."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Dict, List
from google.longrunning.operations_pb2 import Operation
from google.cloud.discoveryengine import (
    TargetSite,
    SiteSearchEngineServiceClient,
    ListTargetSitesRequest,
    GetTargetSiteRequest,
    CreateTargetSiteRequest,
    DeleteTargetSiteRequest,
    EnableAdvancedSiteSearchRequest,
    DisableAdvancedSiteSearchRequest,
    RecrawlUrisRequest
)

from dfcx_scrapi.core import scrapi_base

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class Sites(scrapi_base.ScrapiBase):
    """Site Search Class for Data Stores in Vertex Search and Conversation."""
    def __init__(
        self,
        project_id: str,
        creds_path: str = None,
        creds_dict: Dict = None,
        creds=None,
        scope=False,
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            scope=scope,
        )

        self.project_id = project_id

    @staticmethod
    def __get_site_type(include_site: bool) -> TargetSite.Type:
        """Get the Site Type mapping."""
        site_type_map = {True: 1, False: 2}

        return site_type_map[include_site]

    def __build_site_search_parent(self, data_store_id: str) -> str:
        """Build the Parent ID needed for Site Search API calls."""
        parts = self._parse_resource_path("data_store", data_store_id)
        location = parts.get("location")
        data_store_id_short = parts.get("data_store")

        return (f"{self._build_data_store_parent(location)}/dataStores/"
                f"{data_store_id_short}/siteSearchEngine")

    def get_sites_map(
            self, data_store_id: str, reverse: bool = False) -> Dict[str, str]:
        """Get a user friendly mapping for Site Names and IDs."""
        if reverse:
            site_dict = {
                site.generated_uri_pattern: site.name
                for site in self.list_sites(data_store_id=data_store_id)
            }

        else:
            site_dict = {
                site.name: site.generated_uri_pattern
                for site in self.list_sites(data_store_id=data_store_id)
            }

        return site_dict

    def list_sites(self, data_store_id: str) -> List[TargetSite]:
        """List all URL patterns for a given Data Store ID."""
        client_options = self._client_options_discovery_engine(data_store_id)
        client = SiteSearchEngineServiceClient(
            credentials=self.creds, client_options=client_options
        )

        parent = self.__build_site_search_parent(data_store_id)
        request = ListTargetSitesRequest(
            parent=parent,
            page_size=1000
        )

        page_result = client.list_target_sites(request=request)

        sites = []
        for response in page_result:
            sites.append(response)

        return sites

    def get_site(self, site_id: str) -> TargetSite:
        """Get a single Site by specified ID."""
        client_options = self._client_options_discovery_engine(site_id)
        client = SiteSearchEngineServiceClient(
            credentials=self.creds, client_options=client_options
        )
        request = GetTargetSiteRequest(name=site_id)
        response = client.get_target_site(request=request)

        return response

    def create_site(self, data_store_id: str, uri_pattern: str,
                    include_site: bool = True, exact_match: bool = False
                    ) -> Operation:
        """Create a new Target Site in the specified Data Store.

        Args:
          data_store_id: the fully qualified Data Store ID
          uri_pattern: The user provided URI pattern from which the
            generated_uri_pattern is generated.
          include_site: If set to True, the uri pattern will be Included. If set
            to False, the site pattern will be Excluded.
          exact_match: If set to false, a uri_pattern is generated to include
            all pages whose address contains the provided_uri_pattern. If set to
            true, an uri_pattern is generated to try to be an exact match of the
            provided_uri_pattern or just the specific page if the
            provided_uri_pattern is a specific one. provided_uri_pattern is
            always normalized to generate the URI pattern to be used by the
            search engine.
        """
        client_options = self._client_options_discovery_engine(data_store_id)
        client = SiteSearchEngineServiceClient(
            credentials=self.creds, client_options=client_options
        )

        target_site = TargetSite()
        target_site.provided_uri_pattern = uri_pattern
        target_site.type_ = self.__get_site_type(include_site)
        target_site.exact_match = exact_match

        parent = self.__build_site_search_parent(data_store_id)
        request = CreateTargetSiteRequest(
            parent=parent,
            target_site=target_site
            )

        operation = client.create_target_site(request=request)

        return operation

    def delete_site(self, site_id: str) -> Operation:
        """Deletes a TargetSite in a Data Store by the specified ID."""
        client_options = self._client_options_discovery_engine(site_id)
        client = SiteSearchEngineServiceClient(
            credentials=self.creds, client_options=client_options
        )

        request = DeleteTargetSiteRequest(name=site_id)
        operation = client.delete_target_site(request=request)

        return operation

    def enable_advanced_site_search(self, data_store_id: str) -> Operation:
        """Enables Advanced Site Search for the provided Data Store ID."""
        client_options = self._client_options_discovery_engine(data_store_id)
        client = SiteSearchEngineServiceClient(
            credentials=self.creds, client_options=client_options
        )

        parent = self.__build_site_search_parent(data_store_id)
        request = EnableAdvancedSiteSearchRequest(
            site_search_engine=parent
            )

        operation = client.enable_advanced_site_search(request=request)

        return operation

    def disable_advanced_site_search(self, data_store_id: str) -> Operation:
        """Disable Advanced Site Search for the provided Data Store ID."""
        client_options = self._client_options_discovery_engine(data_store_id)
        client = SiteSearchEngineServiceClient(
            credentials=self.creds, client_options=client_options
        )

        parent = self.__build_site_search_parent(data_store_id)
        request = DisableAdvancedSiteSearchRequest(
            site_search_engine=parent
            )

        operation = client.disable_advanced_site_search(request=request)

        return operation

    def recrawl_uris(self, data_store_id: str, uris: List[str]) -> Operation:
        """Recrawl the specified set of URIs for the Data Store."""
        client_options = self._client_options_discovery_engine(data_store_id)
        client = SiteSearchEngineServiceClient(
            credentials=self.creds, client_options=client_options
        )

        parent = self.__build_site_search_parent(data_store_id)
        request = RecrawlUrisRequest(
            site_search_engine=parent, uris=uris
            )

        operation = client.recrawl_uris(request=request)

        return operation

    def get_site_index_status(self, site_id: str) -> str:
        """Get the Site Indexing Status f0r the provided Site ID."""
        site = self.get_site(site_id)

        return site.indexing_status

    def get_site_verification_status(self, site_id: str) -> str:
        """Get the Site Verification Status for the provided Site ID."""
        site = self.get_site(site_id)
        state = site.site_verification_info.site_verification_state

        verify_map = {
            1: "VERIFIED: Site ownership verified",
            2: ("UNVERIFIED: Site ownership pending verification or"
                " verification failed"),
            3: ("EXEMPTED: Site exempt from verification, e.g. a public website"
                " that opens to all.")
        }

        return verify_map[state.value]
